
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // if DispDel > 0 then we wait for DispDel cycles to complete before update
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  
  if (DispDel > 0) {DispDel--; return;} // don't update display if DispDel > 0
  if (DispClr) {display.clear(); DispClr = false; return;}
  if (DispDisp) {display.display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each
    if (DispBack) {
      // draw a back arrow on the screen
//      display.clear();
//      display.drawLine(48,32,80,16);
//      display.drawLine(48,32,80,48);
//      display.drawLine(80,16,80,48);
//      DispCnt = 10;
    } else {
      if (DispNow > 0) {
        // special display modes used in calibration etc.
        switch(DispNow) {
//          case 1: DispCnt = 3; Display_ServoVals(); break;
        }
      } else {
        if (TEST) {
          // in TEST mode we use a special display
          Display_TESTvalues();
        } else {
          // normal menu screens, display depends on menu modes
          switch(DispMode) {
            case 0: Display_Battery(); break;
            case 1: Display_Range(); break;
            case 2: Display_Walk_Fwd(); break;
            case 3: Display_Walk_Bck(); break;
            case 4: Display_Autonomous(); break;
            case 5: Display_Neutral_Left(); break;
            case 6: Display_Neutral_Right(); break;
          }
        }
      }
    }
    // the following line puts a rectangle around every display, showing the boundry region
    if (Border) {display.drawRect(0,0,128,64);}
  }
}

// --------------------------------------------------------------------------------

void Display_Autonomous() {
  // display Autonomous data
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 2, "Autonomous");
  display.drawString(64,24, "Range " + String(Range));
  switch(AutoMode) {
    case 0: Any$ = ""; break;
    case 1: Any$ = "Forward"; break;
    case 2: Any$ = "Scanning"; break;
    case 3: Any$ = "Turning"; break;
    case 4: Any$ = "Walking"; break;
  }
  display.drawString(64,46,Any$);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // normal screen to be displayed
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  display.drawString(64, 14, "Battery");
  if (USB) {
    display.drawString(64, 34, "USB Mode");
  } else {
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    display.drawString(64, 34, String(float(BatAvg)/BatCal) + "v  " + String(AnyLong) + "%");
  }
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Battery_Low() {
  // displayed when going into battery low shut-down
  display.setBrightness(25);
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  display.drawString(64,14, "BATTERY");
  display.drawString(64,34, "LOW");
  display.display();
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
//  display.flipScreenVertically();
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 6,24,4,24,6
  display.drawString(64, 6, "HexBot");
  display.drawString(64, 34, "2");
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Neutral_Left() {
  // display neutral turn left data
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 2, "Turn Left");
  display.drawString(64,24, "Gear " + String(WalkSpeed));
  display.drawString(64,46,String(WalkFreq * 4));
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Neutral_Right() {
  // display neutral turn right data
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 2, "Turn Right");
  display.drawString(64,24, "Gear " + String(WalkSpeed));
  display.drawString(64,46,String(WalkFreq * 4));
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_PWM() {
  // display servoAny
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 6,24,4,24,6
  int16_t zY = 20;
  if (WiFiRxRec) {
    display.drawString(64, 6, "WiFi"); zY = 34;
  }
  display.drawString(64, zY, "PWM " + String(ServoAny));
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 5;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Range() {
  // display VL53L1X range measurement
  String zSL = ""; String zSR = "";
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
       if (MainTask == 2) {display.drawString(64, 2, "Back Away");}
  else if (MainTask == 3) {display.drawString(64, 2, "Track Mode");}
  else {display.drawString(64, 2, "Range");}
  if (RangeEn) {
    if (RangeUL) {zSL = "< "; zSR = " >";}
    if (RangeLL) {zSL = "> "; zSR = " <";}
    display.drawString(64,24, zSL + String(RangeRaw) + " mm" + zSR);
  } else {display.drawString(65,24, "---");}
  if (LTOFcalc) {
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0,46,String(LTOFfps) + " fps");
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    if ((LTOFspd10 == 0) || RangeUL) {
      display.drawString(127,46, "--- m/s");
    } else {
      display.drawString(127,46, String(float(LTOFspd10)/10.0,1) + " m/s");
    }
  } else if (HeadActive) {
    // head is moving so display the angle
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0,46,String(LTOFfps) + " fps");
    
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    GetHeadAngle(); display.drawString(127,46,String(HeadAng) + "°");
  } else {display.drawString(64,46,String(LTOFfps) + " fps");}
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_TestMode() {
  // display TEST MODE message
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 6, "TEST");
  display.drawString(64, 34, "MODE");
  display.drawRect(0,0,128,64);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 50;       // wait 2s before displaying next count
}

// --------------------------------------------------------------------------------

void Display_TESTvalues() {
  // used in TEST mode to display test values
  display.drawRect(0,0,128,64);
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  if (USB) {
    display.drawString(2, 2, "Batt: USB");
  } else {
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    display.drawString(2, 2, "Batt: " + String(float(BatAvg)/BatCal) + "v  " + String(AnyLong) + "%");
  }
  if (SerialRx) {
    // receiving serial data in TEST mode so assume this is the 16-channel controller
    display.drawString(2, 21, "PWM: " + String(servoVal[ServoPnt]) + "  S" + String(ServoPnt));
  } else {
    // no serial comms so display PWM associated with TestMoveServos()
    if (TestTask == 1) {display.drawString(2, 21, "PWM: " + String(ServoAny));}
    else {display.drawString(2, 21, "Cal. Sweep " + String(TgtCnt));}
  }
  if (RangeEn) {display.drawString(2, 39, "Rnge: " + String(Range));}
  else {display.drawString(2, 39, "Rnge: ---");}
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 5;        // wait 200 ms before displaying next count
}

// --------------------------------------------------------------------------------

void DisplayText2(String zT0,String zT1) {
  // display two lines of text of equal height
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  display.drawString(64, 14, zT0);
  display.drawString(64, 34, zT1);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void DisplayText2L1(String zT0,String zT1) {
  // display two lines of text, 1st line is large
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 8,24,8,16,8
  display.drawString(64,  8, zT0);
  display.setFont(ArialMT_Plain_16);
  display.drawString(64, 40, zT1);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 25;       // wait 1 sec before display next screen
}

// --------------------------------------------------------------------------------

void Display_Walk_Bck() {
  // display walk forward data
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 2, "Walk Bck");
  display.drawString(64,24, "Gear " + String(WalkSpeed));
  display.drawString(64,46,String(WalkFreq * 4));
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------

void Display_Walk_Fwd() {
  // display walk forward data
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 2, "Walk Fwd");
  display.drawString(64,24, "Gear " + String(WalkSpeed));
  display.drawString(64,46,String(WalkFreq * 4));
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
  DispDel = 3;        // wait 200ms before displaying next count
}

// --------------------------------------------------------------------------------
