
// ---------------------------------------------------------------------------------

void attachHeadServo() {
  // head servo and set its position to HeadVal
  if (ESC || USB) return;
  
  servoHeadEn = true;
  servoHead.setPeriodHertz(PWM_Freq);    // set PWM frequency
  servoHead.attach(HeadPin,500,2400);
  servoHead.writeMicroseconds(HeadVal);
  HeadTgt = HeadVal; // correct target positions
  HeadAtt = true;
  Serial.println("Head servo attached.");
}

// ---------------------------------------------------------------------------------

void attachAServo(int zS) {
  // attach servo zS using servoVal[zS] but don't modify the ServoEn flag
  if (ESC || USB) return;
  
  servoInst[zS].setPeriodHertz(PWM_Freq);    // set PWM frequency
  servoInst[zS].attach(servoPins[zS],500,2400);
  servoInst[zS].writeMicroseconds(servoVal[zS]);
  servoAtt[zS] = true;
  Serial.println("Servos " + String(zS) + " attached.");
}

// ---------------------------------------------------------------------------------

void attachServos(int zDel) {
  // attach all 12 leg servos
  // delay used during power-up reset, otherwiswe set to 0
  if (ESC || USB) return;
  
  ServoEn = true; // flag servos are attached
  // attach the leg servos
  for (int zS = 0; zS <= 11; zS++) {
    servoInst[zS].setPeriodHertz(PWM_Freq);    // set PWM frequency
    servoInst[zS].attach(servoPins[zS],500,2400);
    servoInst[zS].writeMicroseconds(servoVal[zS]);
    servoAtt[zS] = true;
    delay(zDel);
  }
  Serial.println("Leg servos attached.");
}

// --------------------------------------------------------------------------------

void battRead() {
  // read the analogue voltage connected to the battery every 20ms
  BatVol = analogRead(BatPin);
//  BatVol-= BatDist; BatDist++;  // uncomment this line to simulate rapid battery discharge
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/50;     // ADC is averaged over 50 readings to remove noise
  if (USB) {return;}

  // detect critical battery condition and disable HexBot 2
  if (BatAvg <= BatCritical){
    // critical battery voltage reached
    Display_Battery_Low();            // display warning
    FastLED.clear(); FastLED.show();  // turn OFF LEDs
    VL53L1X_OFF();                    // turn OFF LTOF
    GoToStore(50);
    while (TgtCnt > 0) {
      // complete the goto rest from here
      Leg_Engine(); delay(20);
    }
    // and that's it
    while (true) {yield();}
  }
}

// --------------------------------------------------------------------------------

void checkTgts() {
  // ensure that set targets are within servo min/max limits
  servoTgt[ 0] = min(servoTgt[ 0],Max0); servoTgt[ 0] = max(servoTgt[ 0],Min0);
  servoTgt[ 1] = min(servoTgt[ 1],Max1); servoTgt[ 1] = max(servoTgt[ 1],Min1);
  servoTgt[ 2] = min(servoTgt[ 2],Max2); servoTgt[ 2] = max(servoTgt[ 2],Min2);
  servoTgt[ 3] = min(servoTgt[ 3],Max3); servoTgt[ 3] = max(servoTgt[ 3],Min3);
  servoTgt[ 4] = min(servoTgt[ 4],Max4); servoTgt[ 4] = max(servoTgt[ 4],Min4);
  servoTgt[ 5] = min(servoTgt[ 5],Max5); servoTgt[ 5] = max(servoTgt[ 5],Min5);
  servoTgt[ 6] = min(servoTgt[ 6],Max6); servoTgt[ 6] = max(servoTgt[ 6],Min6);
  servoTgt[ 7] = min(servoTgt[ 7],Max7); servoTgt[ 7] = max(servoTgt[ 7],Min7);
  servoTgt[ 8] = min(servoTgt[ 8],Max8); servoTgt[ 8] = max(servoTgt[ 8],Min8);
  servoTgt[ 9] = min(servoTgt[ 9],Max9); servoTgt[ 9] = max(servoTgt[ 9],Min9);
  servoTgt[10] = min(servoTgt[10],MaxA); servoTgt[10] = max(servoTgt[10],MinA);
  servoTgt[11] = min(servoTgt[11],MaxB); servoTgt[11] = max(servoTgt[11],MinB);
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 10ms later, just in case
  // there is another packet behind it coming from the controller
  next40ms = millis() + 10;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;          // command terminator
//    case '+': cmdVal = -1; return;      // set flag for ML+. command
    case '-': cmdSgn = -1; return;      // numeric sign
//    case '!': SetMainTask(98); return;  // return to 'Rest' position
//    case '#': Ping = 50; AutoOffTimer = 0;  LEDTaskPnt = 3; return;   // 'ping' sent from application, disables auto-rest function
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    switch (keyChar) {
      // check for valid mode and convert lower-case
      case 's': cmdMode = 'S'; break;
      case 'S': cmdMode = 'S'; break;
      case 'v': cmdMode = 'V'; break;
      case 'V': cmdMode = 'V'; break;
      case 'w': cmdMode = 'W'; break;
      case 'W': cmdMode = 'W'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'a': cmdType = 'A'; break;
      case 'c': cmdType = 'C'; break;
      case 'd': cmdType = 'D'; break;
      case 'h': cmdType = 'H'; break;
      case 'l': cmdType = 'L'; break;
      case 'o': cmdType = 'O'; break;
      case 'r': cmdType = 'R'; break;
      case 's': cmdType = 'S'; break;
      case 't': cmdType = 'T'; break;
      case 'u': cmdType = 'U'; break;
      case 'v': cmdType = 'V'; break;
      case 'x': cmdType = 'X'; break;
      case 'y': cmdType = 'Y'; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void detachHeadServo() {
  // detach head servo
  // as the micro servo uses GPIO2, which is also connected to the buitlin LED, we
  // need to pull this pin LOW when not attached to the servo
  if (!servoHeadEn) {return;} // already detached
  
  unsigned long zMillis = millis();
  servoHeadEn = false;  // flag servo is detached
  while (digitalRead(HeadPin) == LOW) {}
  // was HIGH or just gone HIGH, so wait for it to go LOW
  while (digitalRead(HeadPin) == HIGH) {}
  // just gone LOW so detach and pull pin LOW
  servoHead.detach();
  pinMode(HeadPin,OUTPUT); digitalWrite(HeadPin,LOW);
  HeadAtt = false;
  zMillis = millis() - zMillis;
  synchLoopTimers(zMillis); // re-synch loop timers
  Serial.println("Head servo detached within " + String(zMillis) + "ms");
}

// ---------------------------------------------------------------------------------

void detachLegServos() {
  // detatch all of the leg servos, normally used when at rest
  // servo pins PWM are monitored to detach when signal is LOW
  if (!ServoEn) {return;}
  
  unsigned long zMillis = millis();
  ServoEn = false; // flag servos are detached
  // detach the leg servos
  for (int zS = 0; zS <= 11; zS++) {
    if (servoAtt[zS]) {
      // start by looking for a LOW in teh PWM signal
      while (digitalRead(servoPins[zS]) == LOW) {}
      // was HIGH or just gone HIGH, so wait for it to go LOW
      while (digitalRead(servoPins[zS]) == HIGH) {}
      // just gone LOW so detach and pull pin LOW
      servoInst[zS].detach();
      pinMode(servoPins[zS],OUTPUT); digitalWrite(servoPins[zS],LOW);
    }
    servoAtt[zS] = false;
  }
  zMillis = millis() - zMillis;
  synchLoopTimers(zMillis); // re-synch loop timers
  Serial.println("Leg servos detached within " + String(zMillis) + "ms");
}

// ---------------------------------------------------------------------------------

void disableWiFi() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  // SAnn.  - set target servo angle in degrees
  // SD.    - detach the current target servo pin
  // SLnn.  - set target servo lower limit in microseconds
  // SRn.   - set the state of the echo flag, 0 = OFF, 1 = ON
  // STnn.  - set the target servo number 0 - 16, but don't feedback value
  // SUnn.  - set target servo upper limit in microseconds
  // SVnn.  - set target servo angle in microseconds
  // V?.    - report VL53L1X status
  // VCnn.  - set centre of VL53L1X ROI using table number
  // VOn.   - turns VL53L1X n = 1 ON or n = 0 OFF
  // VXnn.  - set VL53L1X ROI width as nn, used with VCnn. Default = 16
  // VYnn.  - set VL53L1X ROI height as nn, used with VCnn. Default = 16
  // W?.    - report current walk values
  // WDnn.  - set walk leg down angle
  // WHnn.  - set walk leg up angle
  // WLnn.  - set walk sequence lower angle limit
  // WSnn.  - set start angle for all legs.
  // WUnn.  - set walk sequence upper angle limit
  // WX.    - execute/stop a walks sequence

  cmdVal*= cmdSgn;  // correct cmdVal with the sign value
  
  switch (cmdMode) {
    case ' ': break;
    case 'S':
      // keyboard/serial commands used to set servo values
      switch (cmdType) {
        case 'A': // set target servo angle in degrees
          SetAngle(ServoPnt,cmdVal); break;
        case 'D': // detach the current target servo pin
          if (ServoPnt < 12) {servoAtt[ServoPnt] = false; servoInst[ServoPnt].detach();}
          else {HeadAtt = false; servoHead.detach();}
          break;
        case 'L': // set target servo lower limit in microseconds
          if (ServoPnt < 12) {servoMin[ServoPnt] = cmdVal;}
          else {HeadMin = cmdVal;}
          break;
        case 'R': // set the state of the Echo flag
          if (cmdVal > 0) {Echo = true; PrintTx += "Echo ON\n";} else {Echo = false;}
          break;
        case 'T': // set target servo pointer in range 0 - 12
          if (cmdVal < 0) {cmdVal = ServoPnt;}
          if (cmdVal > 12) {cmdVal = ServoPnt;}
          ServoPnt = cmdVal; break;
        case 'U': // set target servo upper limit in microseconds
          if (ServoPnt < 12) {servoMax[ServoPnt] = cmdVal;}
          else {HeadMax = cmdVal;}
          break;
        case 'V': // set target servo angle in microseconds
          if (ServoPnt < 12) {
            // ServoPnt is pointing at a leg servo
            servoVal[ServoPnt] = cmdVal;
            if (!servoAtt[ServoPnt]) {
              // if not attached, attach it
              servoAtt[ServoPnt] = true;
              servoInst[ServoPnt].setPeriodHertz(PWM_Freq);    // set PWM frequency
              servoInst[ServoPnt].attach(servoPins[ServoPnt],500,2400);
            }
            servoInst[ServoPnt].writeMicroseconds(servoVal[ServoPnt]);
          } else {
            // ServoPnt is pointing at the head servo
            HeadVal = cmdVal;
            if (!HeadAtt) {
              // if not attached, attach it
              HeadAtt = true;
              servoHead.setPeriodHertz(PWM_Freq);
              servoHead.attach(HeadPin,500,2400);
            }
            servoHead.writeMicroseconds(HeadVal);
          }
          break;
      } break;
    case 'V':
      // VL53L1X commands
      switch (cmdType) {
        case '?': // report VL53L1X status
          Serial.println("VL53L1X Task " + String(VL53L1X_Task));
          Serial.println("VL53L1X OC   " + String(VL53L1X_OC));
          Serial.println("VL53L1X ROIX " + String(VL53L1X_ROIX));
          Serial.println("VL53L1X ROIY " + String(VL53L1X_ROIY));
          break;
        case 'C': // set centre of ROI area and width/height
          cmdVal = min(cmdVal,255); cmdVal = max(cmdVal,0);
          VL53L1X_OC = cmdVal;
          VL53L1X.setROI(VL53L1X_ROIX,VL53L1X_ROIY,VL53L1X_OC);
          Serial.println("VC" + String(VL53L1X_OC) + ".");
          break;
        case 'O': // turn VL53L1X sensor ON/OFF
          if (cmdVal == 0) {VL53L1X_ON();} else {VL53L1X_OFF();}
          Serial.println("VO" + String(VL53L1X_OC) + ".");
          break;
        case 'X': // set VL53L1X ROI width
          cmdVal = min(cmdVal,16); cmdVal = max(cmdVal,0);
          VL53L1X_ROIX = cmdVal;
          Serial.println("VX" + String(VL53L1X_ROIX) + ".");
          break;
        case 'Y': // set VL53L1X ROI height
          cmdVal = min(cmdVal,16); cmdVal = max(cmdVal,0);
          VL53L1X_ROIY = cmdVal;
          Serial.println("VY" + String(VL53L1X_ROIY) + ".");
          break;
      }
      break;
    case 'W':
      // keyboard/serial commands used to set direct walk values
      switch (cmdType) {
        case '?': // report values
          ReportWalkVals(); break;
        case 'D': // set walk W_Dn value
          cmdVal = min(cmdVal,130); cmdVal = max(cmdVal,90);
          if (cmdVal < W_Up) {W_Dn = cmdVal; Serial.println("WD" + String(W_Dn) + ".");}
          break;
        case 'H': // set walk W_Up value
          cmdVal = min(cmdVal,130); cmdVal = max(cmdVal,90);
          if (cmdVal > W_Dn) {W_Up = cmdVal; Serial.println("WH" + String(W_Up) + ".");}
          break;
        case 'L': // set walk LL value
          cmdVal = (cmdVal/2) * 2;  // ensure it is always divisible by 2
          cmdVal = min(cmdVal,120); cmdVal = max(cmdVal,70);
          if (cmdVal < Walk_U) {Walk_L = cmdVal; Serial.println("WL" + String(Walk_L) + ".");}
          break;
        case 'S': // set walk start value
          cmdVal = min(cmdVal,110); cmdVal = max(cmdVal,70);
          if (cmdVal < Walk_L) {Walk_L = (cmdVal/2) * 2;}
          if (cmdVal > Walk_U) {Walk_U = (cmdVal/2) * 2;}
          Walk_S = cmdVal; Serial.println("WS" + String(Walk_S) + ".");
          break;
        case 'U': // set walk UL value
          cmdVal = (cmdVal/2) * 2;  // ensure it is always divisible by 2
          cmdVal = min(cmdVal,120); cmdVal = max(cmdVal,70);
          if (cmdVal > Walk_L) {Walk_U = cmdVal; Serial.println("WU" + String(Walk_U) + ".");}
          break;
        case 'X': // execute/end a walk process
          if (Walk > 0) {Walk = 0;} else {Walk = 98;}
          break;
      } break;
      break;
    default:
      PrintTx += "Unknown cmd: " + cmdMode + cmdType + String(cmdVal) + "\n";
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// ---------------------------------------------------------------------------------

void flashWhilstDown(int zON,int zOFF) {
  // flash LEDs whilst waiting for button switch to be released
//  while (analogRead(A6) <= 16) {
//    digitalWrite(LED0, LOW); digitalWrite(LED4, LOW);
//    delay(zON);
//    digitalWrite(LED0, HIGH); digitalWrite(LED4, HIGH);
//    if (analogRead(A6) > 16) {break;}
//    delay(zOFF);
//  }
}

// --------------------------------------------------------------------------------

void GetRndCol(byte zB) {
  // return a random colour of strength zB
  if (zB < 2) {zB = 2;}
  switch(random(6)) {
    case 0: Col_R =   zB; Col_G =    0; Col_B =    0; break;  // red
    case 1: Col_R = zB/2; Col_G = zB/2; Col_B =    0; break;  // yellow
    case 2: Col_R =    0; Col_G =   zB; Col_B =    0; break;  // green
    case 3: Col_R =    0; Col_G = zB/2; Col_B = zB/2; break;  // cyan
    case 4: Col_R =    0; Col_G =    0; Col_B =   zB; break;  // blue
    case 5: Col_R = zB/2; Col_G =    0; Col_B = zB/2; break;  // magenta
  }
}

// --------------------------------------------------------------------------------

void LED_Main() {
  // tasks for manipulating RGB LED patterns
  // Blink clock and modes
  Blink = false; BlinkCnt++; if (BlinkCnt > 30) {BlinkCnt = 0; Blink = true;}
  // switch over-rides
  if ((sw0State == LOW) || (sw1State == LOW)) {
    FastLED.clear();
    if (sw0State == LOW) {LED_SW0();}  // force left side LEDs
    if (sw1State == LOW) {LED_SW1();}  // force right side LEDs
    LED_Run = false; LEDshow = true; LED_Del = 1; return;
  }
  if (LED_Del > 0) {LED_Del--; LED_Run = false; return;} // delay by skipping tasks
  
  if (TEST) {LED_Test(); return;} // TEST mode over-ride

//  Serial.println(LedMode);
  switch(LedMode) {
    case 0: // doing nothing
      LED_Store(); break;
    case 1: // just standing 60/120
      LED_Stand(); break;
    case 2: // standing ready 90/90
      LED_Ready(); break;
    case 5: // LTOF range
      LED_Range(); break;
    case 10: // random LED colours
      GetRndCol(2); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(4); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      GetRndCol(8); LED[random(13)].setRGB(Col_R,Col_G,Col_B);
      LED_Run = false; LEDshow = true; LED_Del = 0;
      break;
    case 21: // walking forward
      LED_WalkFwd(); break;
    case 23: // walking backward
      LED_WalkBck(); break;
    case 25: // walking neutral turn left
      LED_NeutLft(); break;
    case 26: // walking neutral turn right
      LED_NeutRht(); break;
  }
}

// --------------------------------------------------------------------------------

void LED_NeutLft() {
  // generates a neutral turn left pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[ 8].setRGB(3,3,0); LED[7].setRGB(3,3,0);
      LED[ 9].setRGB(0,0,4); LED[6].setRGB(0,0,4);
      LED[10].setRGB(0,0,2); LED[5].setRGB(0,0,2);
      LED[11].setRGB(0,0,1); LED[4].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LedCnt++;
        if (LedCnt >= 4) {
          LED[0] = LED[8]; LED[1] = LED[9]; LED[2] = LED[10]; LED[3] = LED[11];
          LED[4] = LED[8]; LED[5] = LED[9]; LED[6] = LED[10]; LED[7] = LED[11];
          LEDX.r = LED[8].r; LEDX.g = LED[8].g; LEDX.b = LED[8].b; // grab the front left LED rgb values
          LED[8] = LED[9]; LED[9] = LED[10]; LED[10] = LED[11]; LED[11] = LED[12];
          LED[12].r = LEDX.r; LED[12].g = LEDX.g;  LED[12].b = LEDX.b;
          LED_Run = false; LEDshow = true; LedCnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_NeutRht() {
  // generates a neutral turn right pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[12].setRGB(3,3,0); LED[7].setRGB(3,3,0);
      LED[11].setRGB(0,0,4); LED[6].setRGB(0,0,4);
      LED[10].setRGB(0,0,2); LED[5].setRGB(0,0,2);
      LED[ 9].setRGB(0,0,1); LED[4].setRGB(0,0,1);
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LedCnt++;
        if (LedCnt >= 4) {
          LED[3] = LED[12]; LED[2] = LED[11]; LED[1] = LED[10]; LED[0] = LED[9];
          LED[7] = LED[12]; LED[6] = LED[11]; LED[5] = LED[10]; LED[4] = LED[9];
          LEDX.r = LED[12].r; LEDX.g = LED[12].g; LEDX.b = LED[12].b; // grab the front right LED rgb values
          LED[12] = LED[11]; LED[11] = LED[10]; LED[10] = LED[9]; LED[9] = LED[8];
          LED[8].r = LEDX.r; LED[8].g = LEDX.g;  LED[8].b = LEDX.b;
          LED_Run = false; LEDshow = true; LedCnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_Range() {
  // drive LEDs in response to LTOF ranging   
  FastLED.clear();
  if (RangeRaw <= RangeMax) {
    anyVal = map(Range,RangeMin,RangeMax,110,0);
    if (anyVal >=  10) {LED[ 8].setRGB(1,1,0);}
    if (anyVal >=  20) {LED[ 8].setRGB(5,0,0);}
    if (anyVal >=  30) {LED[ 9].setRGB(1,1,0); LED[ 0].setRGB(1,1,0);  LED[ 7].setRGB(1,1,0);}
    if (anyVal >=  40) {LED[ 9].setRGB(5,0,0); LED[ 0].setRGB(5,0,0);  LED[ 7].setRGB(5,0,0);}
    if (anyVal >=  50) {LED[10].setRGB(1,1,0); LED[ 1].setRGB(1,1,0);  LED[ 6].setRGB(1,1,0);}
    if (anyVal >=  60) {LED[10].setRGB(5,0,0); LED[ 1].setRGB(5,0,0);  LED[ 6].setRGB(5,0,0);}
    if (anyVal >=  70) {LED[11].setRGB(1,1,0); LED[ 2].setRGB(1,1,0);  LED[ 5].setRGB(1,1,0);}
    if (anyVal >=  80) {LED[11].setRGB(5,0,0); LED[ 2].setRGB(5,0,0);  LED[ 5].setRGB(5,0,0);}
    if (anyVal >=  90) {LED[12].setRGB(1,1,0); LED[ 3].setRGB(1,1,0);  LED[ 4].setRGB(1,1,0);}
    if (anyVal >= 100) {LED[12].setRGB(5,0,0); LED[ 3].setRGB(5,0,0);  LED[ 4].setRGB(5,0,0);}
  } else {
    if (Blink) {LED[ 8].setRGB(1,0,0); LED[ 9].setRGB(1,0,0); LED[10].setRGB(1,0,0); LED[11].setRGB(1,0,0); LED[12].setRGB(1,0,0);}
  }
  LED_Run = false; LEDshow = true; LED_Del = 0;
}

// --------------------------------------------------------------------------------

void LED_Ready() {
  // Red walking dots   
  FastLED.clear();
  switch(LedCnt) {
    case 0: LED[10].setRGB(2,0,0); break;
    case 1: LED[ 9].setRGB(2,0,0); LED[11].setRGB(2,0,0); break;
    case 2: LED[ 8].setRGB(2,0,0); LED[12].setRGB(2,0,0); break;
    case 3: LED[ 0].setRGB(2,0,0); LED[ 7].setRGB(2,0,0); break;
    case 4: LED[ 1].setRGB(2,0,0); LED[ 6].setRGB(2,0,0); break;
    case 5: LED[ 2].setRGB(2,0,0); LED[ 5].setRGB(2,0,0); break;
    case 6: LED[ 3].setRGB(2,0,0); LED[ 4].setRGB(2,0,0); break;
    default:
      LedCnt = -1; break;
  } LedCnt++;
  LED_Run = false; LEDshow = true; LED_Del = 3;
}

// --------------------------------------------------------------------------------

void LED_Set_All(byte zR,byte zG,byte zB) {
  // sets all of the LEDs to one colour immediately
  for (byte zL = 0; zL < 13;zL++) {LED[zL].setRGB(zR,zG,zB);}
  FastLED.show();
}

// --------------------------------------------------------------------------------

void LED_Stand() {
  // Purple walking dots when servos are powered, otherwise green
  if (!ServoEn) {LED_Store(); return;}
  FastLED.clear();
  switch(LedCnt) {
    case 0: LED[10].setRGB(1,0,1); break;
    case 1: LED[ 9].setRGB(1,0,1); LED[11].setRGB(1,0,1); break;
    case 2: LED[ 8].setRGB(1,0,1); LED[12].setRGB(1,0,1); break;
    case 3: LED[ 0].setRGB(1,0,1); LED[ 7].setRGB(1,0,1); break;
    case 4: LED[ 1].setRGB(1,0,1); LED[ 6].setRGB(1,0,1); break;
    case 5: LED[ 2].setRGB(1,0,1); LED[ 5].setRGB(1,0,1); break;
    case 6: LED[ 3].setRGB(1,0,1); LED[ 4].setRGB(1,0,1); break;
    default:
      LedCnt = -1; break;
  } LedCnt++;
  LED_Run = false; LEDshow = true; LED_Del = 6;
}

// --------------------------------------------------------------------------------

void LED_Store() {
  // default, slow green walking dots   
  FastLED.clear();
  switch(LedCnt) {
    case 0: LED[10].setRGB(0,1,0); break;
    case 1: LED[ 9].setRGB(0,1,0); LED[11].setRGB(0,1,0); break;
    case 2: LED[ 8].setRGB(0,1,0); LED[12].setRGB(0,1,0); break;
    case 3: LED[ 0].setRGB(0,1,0); LED[ 7].setRGB(0,1,0); break;
    case 4: LED[ 1].setRGB(0,1,0); LED[ 6].setRGB(0,1,0); break;
    case 5: LED[ 2].setRGB(0,1,0); LED[ 5].setRGB(0,1,0); break;
    case 6: LED[ 3].setRGB(0,1,0); LED[ 4].setRGB(0,1,0); break;
    default:
      LedCnt = -1; break;
  } LedCnt++;
  LED_Run = false; LEDshow = true; LED_Del = 10;
}

// --------------------------------------------------------------------------------

void LED_SW0() {
  // called as an over-ride when SW0 is pressed
  LED[4].setRGB(0,0,3); LED[5] = LED[4];  LED[6] = LED[5]; LED[7] = LED[6];
}

// --------------------------------------------------------------------------------

void LED_SW1() {
  // called as an over-ride when SW1 is pressed
  LED[0].setRGB(0,0,3); LED[1] = LED[0];  LED[2] = LED[1]; LED[3] = LED[2];
}

// --------------------------------------------------------------------------------

void LED_Test() {
  // scrolling RGB sequence in TEST mode
  // scroll LEDs up
  LED[12] = LED[11]; LED[11] = LED[10]; LED[10] = LED[9]; LED[9] = LED[8];
  LED[8] = LED[7]; LED[7] = LED[6]; LED[6] = LED[5]; LED[5] = LED[4];
  LED[4] = LED[3]; LED[3] = LED[2]; LED[2] = LED[1]; LED[1] = LED[0];
  
//  LED[0].setRGB(0,0,0); // set LED[0] black
  switch(LED_Task) {
    case 0: LED[0].setRGB(2,0,0); break;
    case 3: LED[0].setRGB(0,2,0); break;
    case 6: LED[0].setRGB(0,0,2); break;
  }
  LED_Task++; if (LED_Task > 8) {LED_Task = 0;}
  LED_Run = false; LEDshow = true; LED_Del = 6;
}

// --------------------------------------------------------------------------------

void LED_WalkBck() {
  // generates a walking backward pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[0].setRGB(3,3,0); LED[7].setRGB(3,3,0); // yellow
      LED[1].setRGB(0,0,4); LED[6].setRGB(0,0,4); // bright blue
      LED[2].setRGB(0,0,2); LED[5].setRGB(0,0,2); // mid blue
      LED[3].setRGB(0,0,1); LED[4].setRGB(0,0,1); // min blue
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LedCnt++;
        if (LedCnt >= 4) {
          LEDX.r = LED[10].r; LEDX.g = LED[10].g; LEDX.b = LED[10].b; // grab the front centre LED rgb values
          LED[10] = LED[9]; LED[11] = LED[12]; LED[9] = LED[8]; LED[12] = LED[0]; LED[8] = LED[7];
          LED[0] = LED[1]; LED[7] = LED[6]; LED[1] = LED[2]; LED[6] = LED[5]; LED[2] = LED[3];
          LED[5] = LED[4]; LED[3].r = LEDX.r; LED[3].g = LEDX.g;  LED[3].b = LEDX.b; LED[4] = LED[3];
          LED_Run = false; LEDshow = true; LedCnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void LED_WalkFwd() {
  // generates a walking forward pattern
  switch (LED_Task) {
    case 0: // create LED pattern
      FastLED.clear();
      LED[3].setRGB(3,3,0); LED[4].setRGB(3,3,0); // yellow
      LED[2].setRGB(0,0,4); LED[5].setRGB(0,0,4); // bright blue
      LED[1].setRGB(0,0,2); LED[6].setRGB(0,0,2); // mid blue
      LED[0].setRGB(0,0,1); LED[7].setRGB(0,0,1); // min blue
      LED_Run = false; LEDshow = true; LED_Task++; break;
    case 1: // rotate LED pattern
      if (LedTick) {
        LedTick = false; LedCnt++;
        if (LedCnt >= 4) {
          LEDX.r = LED[10].r; LEDX.g = LED[10].g; LEDX.b = LED[10].b; // grab the front centre LED rgb values
          LED[10] = LED[3]; LED[3] = LED[2]; LED[4] = LED[5]; LED[2] = LED[1]; LED[5] = LED[6];
          LED[1] = LED[0]; LED[6] = LED[7]; LED[0] = LED[12]; LED[7] = LED[8]; LED[12] = LED[11];
          LED[8] = LED[9]; LED[11].r = LEDX.r; LED[11].g = LEDX.g;  LED[11].b = LEDX.b; LED[9] = LED[11];
          LED_Run = false; LEDshow = true; LedCnt = 0;
        }
      } break;
  }
}

// --------------------------------------------------------------------------------

void MemRead() {
  // reads the stored servo values into target values
  for (anyFor = 0;anyFor < 12;anyFor++) {servoTgt[anyFor] = servoMem[anyFor];}
}

// --------------------------------------------------------------------------------

void MemStore() {
  // stores the current servo values in a memory store
  for (anyFor = 0;anyFor < 12;anyFor++) {servoMem[anyFor] = servoVal[anyFor];}
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {SerialRx = true; PrintTgt = 0; decodeKey(keyVal);}
}

// ---------------------------------------------------------------------------------

void readSW0() {
  // read the left button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  sw0State = digitalRead(sw0Pin); // record button state
  if (sw0_Nop) {return;}          // block this function whilst == true
  
  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
  //    Serial.println("SW0 Hi-LO");
      // switch PWM mode ON/OFF
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++; // count on the falling edge
      sw0Timer = 0; // reset the timer
    } else {
      // whilst the button is down adjust the timer at 10ms steps
      // if Wii is not connected test for a long press to invoke demo mode
      sw0DwnTime++; // track button pressed time
      if (TEST) {
        // look for long press >1 sec in TEST mode
        if (sw0DwnTime == 100) {
          if (Testmode == 1) {TestTask = 2;}
          else if (Testmode == 2) {TestTask = 10;} else {TestTask = 2;}
          sw0Cnt = 0; sw0Timer = 0;
        }
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0LastState == LOW) {
      // SW0 has just been released
  //    Serial.println("SW0 Lo-Hi");
    }
    if (sw0Timer > 50) {
      // button released for 1 sec so assume valid button counts
    //    Serial.print(F("swCnt = ")); Serial.println(swCnt);
      if (TEST && (sw0Cnt > 0)) {
        // single button press during test mode
        if (Testmode > 1) {TestTask = 1; ServoInc = 0; ServoAny = 0;}
        else if (ServoInc == 0) {ServoInc = 1;} else {ServoInc = 0;}
      }
      sw0Cnt = 0; sw0Timer = 0; DispCnt = 0;
    }
    sw0DwnTime = 0;
  }
  sw0LastState = sw0State; // record current state
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void readSW1() {
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  // SW1 is read every 10ms
  sw1State = digitalRead(sw1Pin); // record button state
  if (sw1_Nop) {return;}          // block this function whilst == true
  
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
//      Serial.println("SW1 Hi-LO");
      sw1DwnTime = 0; // restart button down time counter
      sw1Cnt++; // count on the falling edge
      sw1Timer = 0; // reset the timer
      if (TEST) {
        // in TEST mode so toggle LTOF ON/OFF
        if (!RangeEn) {VL53L1X_ON();} else {VL53L1X_OFF();}
      } else {
        // not in TEST mode
        if (HexBotActive) {
          // active so respond to SW1 button presses
          String zT$ = "Unknown";
          switch(sw1Cnt) {
            case 1: zT$ = "LTOF Range"; break;
            case 2: zT$ = "Back Away"; break;
            case 3: zT$ = "Track Mode"; break;
            case 4: zT$ = "Autonomous"; break;
          }
          DisplayText2L1(String(sw1Cnt),zT$); DispDel = 50; // extend display period
        } else {
          // inactive, but button press will wake HexBot up
          sw1Cnt == 1;  // prevent count from exceeding 1
        }
      }
    } else {
      sw1DwnTime++; // track button pressed time
      if(sw1Timer >= 200) {
        // button held down for >= 2 seconds
        sw1Cnt = 0; sw1Timer = 0; LedMode = 0;
        if (!HexBotActive) {
          // robot was inactive so stand it up and enable head scanning
          HeadActive = true; 
          DisplayText2("Waking","Up!");
          flashWhilstDown(20,150);
          SetMainTask(1); // go to default state
    //      Serial.println(F("Standing Up!"));
          HexBotActive = true; synchLoopTimers(0); autoOFF = 10;
        } else {
          // robot was active so return it to the power-off state
          DisplayText2("Going To","Sleep..."); DispMode = 0;
          flashWhilstDown(20,300);
          SetMainTask(0); // go to null state
          synchLoopTimers(0);
          GoToStore(50); // go to the resting position within 1 second
    //      Serial.println(F("Going To Rest..."));
    //      loopWhiletgtCnt();
          HexBotActive = false; VL53L1X_OFF(); HeadActive = false;
        }
      }
    }
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1LastState == LOW) {
      // SW1 has just been released
//      Serial.println("SW1 Lo-Hi");
    }
    if (!TEST) {
      // not in TEST mode so set mode based on count value
      if (sw1Timer > 100) {
  //        PrintTx += "sw1Timer = " + String(sw1Timer) + "\n";
        if (sw1Cnt == 1) {
          // goto default state
          if (!HexBotActive) {DisplayText2("Waking","Up!");}
          SetMainTask(1); HexBotActive = true;
        } else if (HexBotActive) {
          // only respond to these counts if stood up and active
          // swCnt == 2 backaway from target mode
          // swCnt == 3 track wall target mode
          // swCnt == 4 autonomous discovery mode
          SetMainTask(sw1Cnt);
        }
        sw1Cnt = 0; sw1Timer = 0; sw1DwnTime= 0; DispCnt = 0;
      }
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // called from the main loop every cycle to check for ESP-NOW data  
  // a state machine is used to separate the data blocks by header type
//  DispDel = 500;  // hold off display updates for 10 sec
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! bloick header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // after the header byte there are 6 data bytes + checksum
      // perform XOR on 6 bytes to check for errors
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        // if Nunchuk data strip off the JX,JY and CZ values
        RxRec = true; // indicate a valid frame
        if (WiiType == 'N') {
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
        } else {
          // not Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        WiFiRxErr++;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
      Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary transfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// ---------------------------------------------------------------------------------

void readWii() {
  // called every 20ms, but wireless transceiver only transmits every 40ms
  // considers data received from the Wii Nunchuk controller over WiFi and 
  // manages all movements demanded by the controller
  if (readWiiCall) {return;}
  readWiiCall = true; // prevent re-entry of this function
  
  if (RxRec) {
  //##############################################################################
  //
  //  WiFi Nunchuk data received
  //
  //##############################################################################
    // Wii data received but we may not have enabled WiFi mode yet
    RxRec = false; RTxTimeout = 25; // set timeout to approx 500ms
//    PrintTx += "JX" + String(JoyX) + " JY" + String(JoyY) + " CZ" + String(CZ) + "\n";
//    Serial.print("CZ "); Serial.println(CZ);
    if (!WiFiEn) {
  //##############################################################################
  //
  //  WiFi is currently disabled, listening
  //
  //##############################################################################
      // we have not received enough 'C' buttons to make the system active
      // the 'C' button alone must be held in to activate the Wii WiFi link
      if (CZ == 1) {
        // C button only is being pressed
        if (WiFiCntC < 1) {DisplayText2("Waking","Up!"); DispMode = 0;}
        WiFiCntC += 2;
//        Serial.println("WiFiCnt = " + String(WiFiCntC));
        if (WiFiCntC > 60) {
          // going active, user has held in 'C' button for sufficient time
          Serial.println("WiFiEn = true");
          SetMainTask(0); // go to default state
          HexBotActive = true;; WiFiEn = true; C_Dn = 8;
          LedMode = 1;  // reset LED tasks
//          if (MainTask != 0) {SetMainTask(0);}
//          if (!AtRest) {SetMainTask(98); LEDMthTask = -1;} // move to squat (default)
          // set defaults here
          GoToStand(50);  // go to standing position
          synchLoopTimers(0); autoOFF = 10;
//          Gear = 1;
        } else {
          // still counting C buttons
          LEDVal = WiFiCntC/10; LedMode = 10;
        }
      } else {
        // C button not pressed or released early so check for a inadequate count
        if (WiFiCntC > 0) {
          // reduce the C-button counter, which has to be sustained to enable WiFi mode
          WiFiCntC--; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC < 1) {LedMode = 0;} else {LedMode = 10;}
        }
      }
    } else {
      
  //##############################################################################
  //
  //  WiFi is enabled
  //
  //##############################################################################
      // WiFi enabled so check for power down
      if (CZ == 0) {
        // both C and Z buttons are being pressed so move towards powering down
        if (WiFiCntC > 0) {
          if (WiFiCntC == 64) {DisplayText2("Going To","Sleep..."); DispMode = 0;}
          WiFiCntC -= 2;
  //        Serial.println("WiFiCnt = " + String(WiFiCntC));
          if (WiFiCntC <= 0) {
            // power-down condition satisfied, so go to rest
            Serial.println("WiFi mode disabled!");
            WiFiCntC = 0; disableWiFi();
            GoToStore(50); // go to the resting position within 1 second
            // set defaults here
             HexBotActive = false; LedMode = 0;
          } else {
            // still counting down CZ buttons
            LedMode = 10;
          }
        }
      } else {
        // Check CZ buttons to see if a speed change is wanted
        if ((CZ & 2) == 0) {
          // C button is pressed, assume a speed change is needed
          C_Dn++;
          if (C_Dn == 2) {
            C_Dn = 8;
            if (WalkSpeed < 5) {WalkSpeed++; SetWalkMax(); DisplayText2("Gear",String(WalkSpeed));}
          }
//          Serial.println("WalkSpeed =" + String(WalkSpeed));
        } else {C_Dn = 0;} // C button released
        // Check Z button. It has two functions. Short press change speed, but if
        // held in it changes the mode of the joysticks
        if ((CZ & 1) == 0) {
          // Z button is pressed
          Z_Dn++; if (Z_Dn > 12) {Z_Dn = 50;} // Z_Dn capped at 50, held for >= 2 seconds
//          Serial.print("Z_Dn "); Serial.println(Z_Dn);
        } else {
          // Z button is not pressed
          if ((Z_Dn > 1) && (Z_Dn <= 10)) {
            // held down for less than 400 ms @ 40ms Rx cycle
            if (WalkSpeed > 1) {WalkSpeed--; SetWalkMax(); DisplayText2("Gear",String(WalkSpeed));}
          } Z_Dn = 0; // Z button released so clear down counter
        }
//        Serial.print("Z_Dn "); Serial.println(Z_Dn);

        if (WiFiCntC < 60) {
          // C-button counter is not at full strength, assume WiFi off released prematurely
          WiFiCntC++; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC == 60) {LED_Task = -1;} else {LED_Task = -10;}
        }
      }
    }
      
  //##############################################################################
  //
  //  Wii joystick demands
  //
  //##############################################################################
    // joystick movements are ignored if:
    // HexBot is not active
    // Wii Joystick has not sent data
    // a MainTask is running
    if ((HexBotActive) && (JoyActive) && (MainTask == 0)) {
      // read the X,Y joysticks and walk if demanded
      JoyActive = false; // prevent stack overflow during moves that call loop()
      // respond to joystick demands every 40ms (25Hz)
      // note direction of travel will only change once stick has been centred
      if ((JoyX >= DbLLX) && (JoyX <= DbULX) && (JoyY >= DbLLY) && (JoyY <= DbULY)) {
        // joystick is in centre deadband position
        JoyMode = 0; // reset the direction of travel mode
//          Serial.print(F("#"));
        if (Walk != 0) {
//          Serial.println(F("Stopping"));
          DispMode = 0; // display battery
          walkInterval = 20000; SetLegsDown(); SetLedMode(2);
          // stop walking if active, but remember direction
          WalkLast = Walk; Walk = 0;
        }
      } else if (JoyMode == 0) {
        autoRest = 0; // reset user time-out 
        if (Z_Dn < 1) {
          // 'Z' button is not pressed so normal walking movements
          if (JoyY > 150) {JoyMode = 1;} // set direction of travel as forward
          else if (JoyY < 106) {JoyMode = -1;} // set direction of travel as backward
          else if (JoyX < 105) {JoyMode = 3;} // set direction of travel as turn left
          else if (JoyX > 149) {JoyMode = 2;} // set direction of travel as turn right
        } else {
          // 'Z' button pressed so perform special actions
          if (JoyX < 105) {JoyMode = 5;} // set direction of travel as walk to left
          else if (JoyX > 149) {JoyMode = 4;} // set direction of travel as walk to right
          else if (JoyY > 149) {JoyMode = 6;} // perform a bow
          else if (JoyY < 106) {JoyMode = 7;} // perform a hello wave
        }
      } else {
        autoRest = 0; // reset user time-out 
        switch (JoyMode) {
          case -1: JoyMoveBwd(); break;
          case 1: JoyMoveFwd(); break;
          case 2: JoyTurnRgt(); break;
          case 3: JoyTurnLft(); break;
          
          case 4: JoyMode = 0; DemoMove_NeutralRgt(); break;
          case 5: JoyMode = 0; DemoMove_NeutralLft(); break;
          case 6: JoyMode = 0; DemoMove_Bow(); break;
          case 7: JoyMode = 0; DemoMove_SitNWave(); break;
        }
      } JoyActive = true; // re-enable this code block
    }
  } else {
  //##############################################################################
  //
  //  No WiFi Nunchuk received recently
  //
  //##############################################################################
    if (WiFiEn) {
      // HexBot is enabled for Wii WiFi control
      // maintain Wii enabled counter
      if ((WiFiCntC < 64) && (CZ != 0)) {
        WiFiCntC++; LEDVal = WiFiCntC/10;
        if (WiFiCntC == 64) {LED_Task = -1;} else {LED_Task = -10;}
      }
    }
  }

  if (WiFiEn) {
  //##############################################################################
  //
  //  WiFi background tasks when WiFi is enabled
  //
  //##############################################################################
    // only do these functions if WiFi is enabled
    if (MoveState == 20) {
      // long un-demanded moves like bowing are watched until they complete
      MoveState = 0; LED_Task = -1;
    }
    
    if (MoveState == -1) {
      // at the end of a joystick move we want to set things back to normal
      LED_Task = -1;
//      SetSpeedDefaults();
      MoveState = 0;
    }
  }
  
  readWiiCall = false; // re-allow calls to this function
//  Serial.print("CntZ "); Serial.println(WiFiCntZ);
//   Serial.println(MoveState);
}

// --------------------------------------------------------------------------------

void ReportWalkVals() {
  // report walk engine values
  Serial.println("\nWalk Engine:");
  Serial.println("WS " + String(Walk_S));
  Serial.println("WL " + String(Walk_L));
  Serial.println("WU " + String(Walk_U));
  Serial.println("WD " + String(W_Dn));
  Serial.println("WH " + String(W_Up));
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void SetAngle(int zS, int zA) {
  // sets a specific servo zS to an angle zA immediately if zS < 20
  // if zS >= 20 then set target only and don't move the servo
  // the mapping function is split either side of 90 for accuracy
  if (ESC) return;
  
  anyVal = 0; if (zS >= 20) {zS -= 20; anyVal = 1;}
  if (zA >= 0) {
    switch (zS) {
      case 0:
        if (zA >= 90) {servoTgt[0] = map(zA,90,223,Ang0_90,Ang0_223);}
        else {servoTgt[0] = map(zA,69,90,Ang0_69,Ang0_90);}
        break;
      case 1:
        if (zA >= 90) {servoTgt[1] = map(zA,90,120,Ang1_90,Ang1_120);}
        else {servoTgt[1] = map(zA,21,90,Ang1_21,Ang1_90);}
        break;
      case 2: 
        if (zA >= 90) {servoTgt[2] = map(zA,90,233,Ang2_90,Ang2_223);}
        else {servoTgt[2] = map(zA,69,90,Ang2_69,Ang2_90);}
        break;
      case 3:
        if (zA >= 90) {servoTgt[3] = map(zA,90,150,Ang3_90,Ang3_150);}
        else {servoTgt[3] = map(zA,30,90,Ang3_30,Ang3_90);}
        break;
      case 4:
        if (zA >= 90) {servoTgt[4] = map(zA,90,180,Ang4_90,Ang4_180);}
        else {servoTgt[4] = map(zA,68,90,Ang4_68,Ang4_90);}
        break;
      case 5:
        if (zA >= 90) {servoTgt[5] = map(zA,90,233,Ang5_90,Ang5_223);}
        else {servoTgt[5] = map(zA,69,90,Ang5_69,Ang5_90);}
        break;
      case 6:
        if (zA >= 90) {servoTgt[6] = map(zA,90,180,Ang6_90,Ang6_180);}
        else {servoTgt[6] = map(zA,68,90,Ang6_68,Ang6_90);}
        break;
      case 7:
        if (zA >= 90) {servoTgt[7] = map(zA,90,223,Ang7_90,Ang7_223);}
        else {servoTgt[7] = map(zA,69,90,Ang7_69,Ang7_90);}
        break;
      case 8:
        if (zA >= 90) {servoTgt[8] = map(zA,90,223,Ang8_90,Ang8_223);}
        else {servoTgt[8] = map(zA,69,90,Ang8_69,Ang8_90);}
        break;
      case 9:
        if (zA >= 90) {servoTgt[9] = map(zA,90,150,Ang9_90,Ang9_150);}
        else {servoTgt[9] = map(zA,30,90,Ang9_30,Ang9_90);}
        break;
      case 10:
        if (zA >= 90) {servoTgt[10] = map(zA,90,223,AngA_90,AngA_223);}
        else {servoTgt[10] = map(zA,69,90,AngA_69,AngA_90);}
        break;
      case 11:
        if (zA >= 90) {servoTgt[11] = map(zA,90,120,AngB_90,AngB_120);}
        else {servoTgt[11] = map(zA,21,90,AngB_21,AngB_90);}
        break;
      case 12:
        if (zA >= 90) {HeadTgt = map(zA,90,160,Head_90,Head_160);}
        else {HeadTgt = map(zA,20,90,Head_20,Head_90);}
        break;
    }
    if (anyVal < 1) {
      // zS is not >= 20 so set servo 0 - 11 immediately
      if (zS < 12) {
        if (!ServoEn) {attachServos(0);}
        if (servoVal[zS] != servoTgt[zS]) {
          servoVal[zS] = servoTgt[zS]; servoInst[zS].writeMicroseconds(servoVal[zS]);
        }
      } else {
        if (!servoHeadEn) {attachHeadServo();}
        if (HeadVal != HeadTgt) {
          HeadVal = HeadTgt; servoHead.writeMicroseconds(HeadVal);
//          Serial.println(HeadVal);
        }
      }
    }
  }
}

// --------------------------------------------------------------------------------

void SetLedMode(int zM) {
  // sets the LED mode and clears the counter
  LedMode = zM; LedCnt = 0; LED_Task = 0; LED_Period = 20;
//  Serial.println("SetLedMode" + String(LedMode));
}

// --------------------------------------------------------------------------------

void SetMainTask(int zMT) {
  // sets the main task and associated flags
  // immediately stop any movement
  TgtCnt = 0; // stop any leg movement
  Walk = 0;   // stop any walking
  if (zMT < 99) {
    // 2 second pause before head becomes active in all active modes
    HeadPause = 100;
  } else {
    // user has pressed the button
    // mode 99 returns all to the standing mode of MainTask 1
    HeadPause = 0; zMT = 0;
    if (HexBotActive) {GoToStand(50);} else {GoToReady(50);}
    autoOFF = 5;
  }
  autoOFF = 5;
  DispMode = 0;         // default battery display
  ESC = false;
  HeadTask = 0;         // effectively turn off head movement tasks
  HeadTaskLast = 0;     // previous head task pointer
  HeadInterval = 20000; // main loop head movement interval in microseconds
  VL53L1X_OFF();        // laser disabled by default
  LTOFcalc = false;     // if == true then calculate speed of ranged object in m/s
  MainRun = false;      // clear MainTask inh
  MainTask = zMT;       // set MainTask
  Range = LtofMax;      // set LTOF range to max
  RangeMax = LtofLimit; // the current limited to Range measurements
  RangeMin = LtofMin;   // the current lower limited to Range measurements
  SetLedMode(0);        // remove any LED flashing
  SubTask = 0;          // reset sub task pointer
  Turn = false;         // default = false; if true turns continuously until = false
  Walked = 0;           // counter used to return robot to a given position
  walkInterval = 20000; synchLoopTimers(0);
  WalkLftDrive = 128;   // drive factor for left-hand side, 128 = max, 0 = min
  WalkRgtDrive = 128;   // drive factor for right-hand side, 128 = max, 0 = min
  if (HeadActive) {HeadToRest();}

  // give VL53L1X 1ms to reset
  delay(1); synchLoopTimers(1);
  
  switch(MainTask) {
    case 1:
      VL53L1X_ON();                   // enable laser ranging task
      DispMode = 1;                   // display range
      if (HeadActive) {HeadTask = 1;  // activate scanning
      } else {LTOFcalc = true;}       // calculate speed of ranged object in m/s
      break;
    case 2:
      VL53L1X_ON();                   // enable back away task
      DispMode = 1;                   // display range
      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
    case 3:
      VL53L1X_ON();                   // enable target tracking task
      DispMode = 1;                   // display range
      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
    case 4:
      VL53L1X_ON();                   // enable autonomous task
      if (HeadActive) {HeadTask = 1;} // activate scanning
      break;
  }
  Serial.println("MainTask = " + String(MainTask));
}

// --------------------------------------------------------------------------------

void SetServoVal(int zS, int zA) {
  // sets a specific servo zS value to an angle zA
  // the mapping function is split either side of 90 for accuracy
  switch (zS) {
    case 0:
      if (zA >= 90) {servoVal[0] = map(zA,90,223,Ang0_90,Ang0_223);}
      else {servoVal[0] = map(zA,69,90,Ang0_69,Ang0_90);}
      break;
    case 1:
      if (zA >= 90) {servoVal[1] = map(zA,90,120,Ang1_90,Ang1_120);}
      else {servoVal[1] = map(zA,21,90,Ang1_21,Ang1_90);}
      break;
    case 2: 
      if (zA >= 90) {servoVal[2] = map(zA,90,233,Ang2_90,Ang2_223);}
      else {servoVal[2] = map(zA,69,90,Ang2_69,Ang2_90);}
      break;
    case 3:
      if (zA >= 90) {servoVal[3] = map(zA,90,150,Ang3_90,Ang3_150);}
      else {servoVal[3] = map(zA,30,90,Ang3_30,Ang3_90);}
      break;
    case 4:
      if (zA >= 90) {servoVal[4] = map(zA,90,180,Ang4_90,Ang4_180);}
      else {servoVal[4] = map(zA,68,90,Ang4_68,Ang4_90);}
      break;
    case 5:
      if (zA >= 90) {servoVal[5] = map(zA,90,233,Ang5_90,Ang5_223);}
      else {servoVal[5] = map(zA,69,90,Ang5_69,Ang5_90);}
      break;
    case 6:
      if (zA >= 90) {servoVal[6] = map(zA,90,180,Ang6_90,Ang6_180);}
      else {servoVal[6] = map(zA,68,90,Ang6_68,Ang6_90);}
      break;
    case 7:
      if (zA >= 90) {servoVal[7] = map(zA,90,223,Ang7_90,Ang7_223);}
      else {servoVal[7] = map(zA,69,90,Ang7_69,Ang7_90);}
      break;
    case 8:
      if (zA >= 90) {servoVal[8] = map(zA,90,223,Ang8_90,Ang8_223);}
      else {servoVal[8] = map(zA,69,90,Ang8_69,Ang8_90);}
      break;
    case 9:
      if (zA >= 90) {servoVal[9] = map(zA,90,150,Ang9_90,Ang9_150);}
      else {servoVal[9] = map(zA,30,90,Ang9_30,Ang9_90);}
      break;
    case 10:
      if (zA >= 90) {servoVal[10] = map(zA,90,223,AngA_90,AngA_223);}
      else {servoVal[10] = map(zA,69,90,AngA_69,AngA_90);}
      break;
    case 11:
      if (zA >= 90) {servoVal[11] = map(zA,90,120,AngB_90,AngB_120);}
      else {servoVal[11] = map(zA,21,90,AngB_21,AngB_90);}
      break;
    case 12:
      if (zA >= 90) {HeadTgt = map(zA,90,160,Head_90,Head_160);}
      else {HeadTgt = map(zA,20,90,Head_20,Head_90);}
      break;
    }
}

// --------------------------------------------------------------------------------

void SetToTgtNow() {
  // set servos to target values immediately, if not at targets
  for (anyFor = 0;anyFor < 12;anyFor++) {
    if (servoVal[anyFor] != servoTgt[anyFor]) {
      servoVal[anyFor] = servoTgt[anyFor]; servoInst[anyFor].writeMicroseconds(servoVal[anyFor]);
    }
  }
}

// --------------------------------------------------------------------------------

void VL53L1X_OFF() {
  // switches OFF the VL53L1X laser ranging device
  digitalWrite(XSHUT_PIN,LOW);  // shut down VL53L1X sensor
  Serial.println("VL53L1X OFF");
  RangeEn = false; VL53L1X_Task = 0;
}

// --------------------------------------------------------------------------------

void VL53L1X_ON() {
  // switches ON the VL53L1X laser ranging device
  // if it was already ON then switch it OFF then ON again
  if (RangeEn) {VL53L1X_OFF(); delay(1); synchLoopTimers(1);}
  
  digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
  if (VL53L1X.init() == false) {
    Serial.println("VL53L1X sensor detected!");
    VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
    VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
    VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
    VL53L1X.setROI(VL53L1X_ROIX,VL53L1X_ROIY,VL53L1X_OC); // Set SPAD ROI
    VL53L1X.startRanging();                 // Starts taking measurements
    Serial.println("VL53L1X ROI SPAD W= " + String(VL53L1X.getROIX()));
    Serial.println("VL53L1X ROI SPAD H= " + String(VL53L1X.getROIY()));
    RangeEn = true; VL53L1X_Task = 1;
    Range = LtofLimit; RangeLast = LtofLimit;
  } else {Serial.println("VL53L1X sensor failed!"); RangeEn = false;}
}

// --------------------------------------------------------------------------------

void VL53L1X_Run() {
  // control the VL53L1X sensor using a series of tasks
  // this code gets timing information, for using the sensor more efficiently
  // once measurements start a skip timer is used to avoid unnecessary readings
  if (millis() < VL53L1X_Skip) {return;}  // allowing for reading delay 32ms
  
  switch(VL53L1X_Task) {
    // case 8 is placed at the front of the list for speed
    case 8: // wait for measurement to complete
      // this case is placed at the head of the list as it will be used most often
      // record previous range just before taking a new measurement
      RangeLast = RangeRaw;
      if (VL53L1X.checkForDataReady()) {
//        t1 = millis() - t1; // how long did the measurement take?
//        Serial.println("VL53L1X.checkForDataReady() in " + String(t1) + "ms");
//        t1 = millis();
        VL53L1X_Skip = millis() + VL53L1X_SkipDel; // set the next measurement delay
        if (VL53L1X.getRangeStatus() == 0) {
          // no ranging errors
//          t0 = micros();
          RangeRaw = VL53L1X.getDistance();
//          t0 = micros() - t0;
//          Serial.println("VL53L1X.getDistance() in " + String(t0) + "µs");
        } else {
          // if an error occurs we set the value to 2 x RangeMax
          RangeRaw = RangeMax + RangeMax;
        }
        LTOFt1 = millis();
//        t0 = micros();
        VL53L1X.clearInterrupt(); // 444µs clears the interrupt and data ready flag
//        t0 = micros() - t0;
//        Serial.println("VL53L1X.clearInterrupt() in " + String(t0) + "µs");
//        Serial.println("");
        // detect and remove transient low values
        if (RangeRaw < RangeMin) {RangeRaw = RangeLast;}
        // range is averaged over two measurements to reduce noise
        Range = (RangeRaw + RangeLast)/2;
        RangeUL = false;
        // limit the range to avoid erroneous values
        if (Range > RangeMax) {Range = RangeMax; RangeUL = true;}
        RangeLL = false;
        if (Range < RangeMin) {Range = RangeMin; RangeLL = true;}
//        Serial.println("Range = " + String(Range) + "mm");
//        Serial.println(Range);
        // calculate the intermeasurement period
        LTOFperiod = LTOFt1 - LTOFt0; LTOFt0 = LTOFt1;
        // calculate the fps frequency
        LTOFfps = 1000/LTOFperiod;
        VL53L1X_Task = 8; // look for another measurement
        if (LTOFcalc) {VL53L1X_Task = 9;}
//        else {Serial.println("0,1000," + String(RangeRaw) + "," + String(Range));}
      } break;
    case 9: // speed calculation is enabled
      // due to noise on the sensor we cannot measure low speed values, and therefore
      // set s threshold of 10mm which is equivalent to 300mm/s minimum
      RangeDiff = abs(RangeRaw - RangeLast); // get the difference
      if (RangeDiff > 10) {
        // movement is > 10mm so calculate speed
        // we apply a peak detector, with 1 sec sustain window
        LTOFspd10 = (RangeDiff * 10)/LTOFperiod;
        if (LTOFspd10 > LTOFspd10L) {LTOFspd10L = LTOFspd10; LTOFspdCnt = 30;}
        else {LTOFspd10 = LTOFspd10L;}
      }
      if (LTOFspdCnt > 0) {
        LTOFspdCnt--;
        if (LTOFspdCnt < 1) {LTOFspd10 = 0; LTOFspd10L = 0;}
      }
//      Serial.println("0,1000," + String(Range) + "," + String(LTOFspd10));
      VL53L1X_Task = 8; // look for another measurement
      break;

    case 1: // switch on sensor
      digitalWrite(XSHUT_PIN,HIGH);  // enable VL53L1X sensor
      VL53L1X_Timer = micros(); 
//      Serial.println("VL53L1X XSHUT HIGH");
      VL53L1X_Task++; break;
    case 2: // check for boot, will time-out after 10ms if sensor does not respond
//      t0 = micros();
      if (VL53L1X.checkBootState()) {
//        t0 = micros() - t0;
        VL53L1X_Timer = micros() - VL53L1X_Timer;
//        Serial.println("VL53L1X.checkBootState() in " + String(t0) + "µs");
        // 825µs needed to boot the device from XSHUT going high
//        Serial.println("VL53L1X Booted! in " + String(VL53L1X_Timer) + "µs");
        VL53L1X_Task++;
      }
      else if (micros() >= (VL53L1X_Timer + 10000)) {VL53L1X_Task = 99;} // time-out
      break;
    case 3: // device booted, so initialise
//      t0 = micros();
      if (VL53L1X.init() == false) {
//        t0 = micros() - t0;
//        Serial.println("VL53L1X.init() in " + String(t0) + "µs"); // 151ms
        VL53L1X_Task++;
      } else {VL53L1X_Task = 99;} // failed to initialise, so turn sensor OFF
      break;
    case 4: // set the distance mode
//      t0 = micros();
      VL53L1X.setDistanceModeShort();         // Sets to 1.3M range
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setDistanceModeShort() in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 5: // set the timing budget as 33ms
//      t0 = micros();
      VL53L1X.setTimingBudgetInMs(33);        // Set timing budget to 33 ms
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setTimingBudgetInMs(33) in " + String(t0) + "µs");
      VL53L1X_Task++; break;
    case 6: // set the distance mode
//      t0 = micros();
      VL53L1X.setIntermeasurementPeriod(33);  // Set measurement period in ms
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.setIntermeasurementPeriod(33) in " + String(t0) + "µs");
      RangeEn = true; VL53L1X_Task++; break;
    case 7: // initiate ranging
//      t0 = micros();
      VL53L1X.startRanging();                 // 446µs Starts taking measurements
//      t0 = micros() - t0;
//      Serial.println("VL53L1X.startRanging() in " + String(t0) + "µs");
//      t1 = millis();
      VL53L1X_Skip = millis() + VL53L1X_SkipDel; // set the initial delay
      LTOFt0 = millis(); VL53L1X_Task++; break;
      
    case 99: // a time-out has occured or we want to turn the sensor OFF
      digitalWrite(XSHUT_PIN,LOW);  // disable VL53L1X sensor
//      Serial.println("VL53L1X XSHUT LOW");
      RangeEn = false; VL53L1X_Task = 0; break;
  }
}

// --------------------------------------------------------------------------------

void Wait_ButtonsDwn() {
  // wait in this function until bothe SW0- and SW1 are released
  while (!digitalRead(sw0Pin) || !digitalRead(sw1Pin)) {
    // one or both buttons held down
    yield();  // let system functions run in background
  }
  sw0Cnt = 0; sw0DwnTime = 0; sw0LastState = HIGH; sw0_Nop = false; sw0State = HIGH; sw0Timer = 0;
  sw1Cnt = 0; sw1DwnTime = 0; sw1LastState = HIGH; sw1_Nop = false; sw1State = HIGH; sw1Timer = 0;
  synchLoopTimers(0);
}

// --------------------------------------------------------------------------------




  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0;     // reset the connection retry counter
  WiFiTx_CB = 1;      // set ready to send flag
  WiFiRxRec = false;  // clear the data received flag
  Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  Tx_Buff.ESPdata[0] = 'H';
  Tx_Buff.ESPdata[1] = 'e';
  Tx_Buff.ESPdata[2] = 'x';
  Tx_Buff.ESPdata[3] = 'B';
  Tx_Buff.ESPdata[4] = 'o';
  Tx_Buff.ESPdata[5] = 't';
  Tx_Buff.ESPdata[6] = '2';
  WiFiTx_len = 7;
//  t0 = micros();
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  if (result == ESP_OK) {
//    Serial.println("Sent with success");
  } else {
//    Serial.println("Error sending the data");
  } 

  WiFiTryNum++; // count the total number of attempts
//  Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------
