
// --------------------------------------------------------------------------------

void CalcHeadAng() {
  // something detected so calculate a steering angle, which is used by
  // the autonomous code
  AnyLong = (512 * (RangeIntR - RangeIntL)) / (RangeIntR + RangeIntL);
  RangeAng = map(AnyLong,-128,128,Head_20,Head_160);
  // now set variables for next half-cycle
  CalcCnt++; // track the number of calculation modes performed
}

// --------------------------------------------------------------------------------

void CheckAutoRest() {
  // check performed every 20ms
  // return to standing position after autoRest time-out
  if (autoRest < autoRestMax) {autoRest++;}
  if (autoRest == 500) {
    // move to the ready position after 5 seconds
//        Serial.println("Ready");
    Serial.println("autoRest == 500");
    MoveToAngles(90,90,90,90,90,90,90,90,90,90,90,90); WalkLast = 0;
    SetLedMode(2);
  } else if (autoRest == 1500) {
    // move to the stand position after 30 seconds of inactivity
//        Serial.println("Standing");
    Serial.println("autoRest == 1500");
    MoveToAngles(90,60,90,90,120,90,120,90,90,90,90,60); // go to the standing position
    SetLedMode(1);
    DetCnt = 100; // detach leg servos 2 sec after movement has stopped
  } else if (autoRest == 15000) {
    // move to the reset store position after 5 minutes of inactivity
//        Serial.println("Sleeping...");
    Serial.println("autoRest == 15000");
    GoToStore(50);
  }
}

// --------------------------------------------------------------------------------

void CheckMinRange() {
  // used in autonomous mode to keep an eye on a target appearing during a
  // half sweep, so this is called at the extremes of the sweep
  if (RangeMinVal <= LtofDet) {
    // something has been detected
    TrgtDet = true;
    // new steering angle is only calculated when something is within range
  } else {
    TrgtDet = false; // nothing has been detected
  }
  if ((Turning) && (RangeMinVal >= LtofStop)) {ESC = true;} // if turning then stop turning
  RangeMinMem = RangeMinVal; RangeMinVal = LtofMax; // store & reset the minimum range
}

// --------------------------------------------------------------------------------

void DemoMove_Bow() {
  // bow movement used to introduce/exit demo move
  Serial.println("DemoMove Bow");
  SetLedMode(21); autoRest = autoRestMax;
  // move to ready position
  // servo    (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
  // type     (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
  MoveToAngles(LegDn,  90,LegDn,  90,  90,LegDn,  90,LegDn,LegDn,  90,LegDn,  90); autoOFF = 0;
  // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
  // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
  GoToAngles(  176,   21,  160,   30,   90,   90,   90,   90,  160,   30,  176,   21,50); // go to the resting position
  delayLoop(1200); // pause
  // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
  // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
  GoToAngles(   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,50); // go to the resting position
  SetLedMode(2); autoRest = 501; WalkLast = 0;
}

// --------------------------------------------------------------------------------

void DemoMove_NeutralLft() {
  // performs a 360° neutral turn to the left
  // useful for checking/calibrating autonomous turns
  Serial.println("DemoMove NeutralLft");
  // servo    (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
  // type     (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
  MoveToAngles(LegDn,  90,LegDn,  90,  90,LegDn,  90,LegDn,LegDn,  90,LegDn,  90); autoOFF = 0;
  SetLedMode(25); autoRest = autoRestMax;
  NeutralTurnLeft(360, 15000); WalkLast = 0;
  // servo    (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
  // type     (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
//  MoveToAngles(LegDn,  90,LegDn,  90,  90,LegDn,  90,LegDn,LegDn,  90,LegDn,  90); autoOFF = 0;
  SetLedMode(2); autoRest = 501;
}

// --------------------------------------------------------------------------------

void DemoMove_NeutralRgt() {
  // performs a 360° neutral turn to the right
  // useful for checking/calibrating autonomous turns
  Serial.println("DemoMove NeutralRgt");
  // servo    (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
  // type     (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
  MoveToAngles(LegDn,  90,LegDn,  90,  90,LegDn,  90,LegDn,LegDn,  90,LegDn,  90); autoOFF = 0;
  SetLedMode(26); autoRest = autoRestMax;
  NeutralTurnRight(360, 15000); WalkLast = 0;
  // servo    (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
  // type     (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
//  MoveToAngles(LegDn,  90,LegDn,  90,  90,LegDn,  90,LegDn,LegDn,  90,LegDn,  90); autoOFF = 0;
  SetLedMode(2); autoRest = 501;
}

// --------------------------------------------------------------------------------

void DemoMove_SitNWave() {
  // sit and wave the right hand
  Serial.println("DemoMove_SitNWave");
  SetLedMode(21); autoRest = autoRestMax;
  // move to legs forward posture
// type         (Shldr, Ang,Claw, Ang)
  MoveLegToAngle(    1,  25,   0,LegDn); // move front right
  MoveLegToAngle(   11,  25,  10,LegDn); // move front left
  MoveLegToAngle(    3,  35,   2,LegDn); // move centre right
  MoveLegToAngle(    9,  35,   8,LegDn); // move centre left
  MoveLegToAngle(    4,  90,   5,LegDn); // move rear right
  MoveLegToAngle(    6,  90,   7,LegDn); // move rear left
  // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
  // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
  GoToAngles(   -1,   -1,   -1,   -1,   -1,  165,   -1,  165,   -1,   -1,   -1,   -1,50); // lower rear legs and bottom
  // waving action
  if (random(2) == 0) {
    // wave with right-hand leg
    // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
    // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
    GoToAngles(  150,   -1,   90,   90,   -1,   -1,   -1,   -1,   90,   90,   75,   90,50); // lower centre legs to lift
    delayLoop(500); // pause
    for (anyFor=0; anyFor<2; anyFor++) {
      // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
      // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
      GoToAngles(  170,   65,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,5); // swing front right leg right
      delayLoop(100); // pause
      // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
      // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
      GoToAngles(  150,   25,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,5); // swing front right leg back
      delayLoop(100); // pause
    }
  } else {
    // wave with left-hand leg
    // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
    // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
    GoToAngles(   75,   90,   90,   90,   -1,   -1,   -1,   -1,   90,   90,  150,   -1,50); // lower centre legs to lift
    delayLoop(500); // pause
    for (anyFor=0; anyFor<2; anyFor++) {
      // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
      // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
      GoToAngles(   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  170,   65,5); // swing front right leg right
      delayLoop(100); // pause
      // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
      // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
      GoToAngles(   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  150,   25,5); // swing front right leg back
      delayLoop(100); // pause
    }
  }
  delayLoop(1000); // pause
  // servo  (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
  // type   ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
  GoToAngles(LegDn,   90,LegDn,   90,   -1,LegDn,   -1,LegDn,LegDn,   90,LegDn,   90,50); // go to the stand position
  SetLedMode(2); autoRest = 501; WalkLast = 0;
}

// --------------------------------------------------------------------------------

void DoHeadTasks() {
  // perform a range of head movement tasks depending on the MainTask
  // in non-autonomous mode the head will look for targets and track them
  // in autonomous mode the head does not seek to find/track a target, it simply
  // scans over a given range to provide steering information
  // HeadTask = 1       initialise head tasks
  // HeadTask = 2 - 4   random head movements
  // HeadTask = 10 - 22 scan a detected target, whilst in range
  // HeadTask = 30 - 37 left/right scanning for autonomous mode, steering and stop
  
  if (HeadTaskRun) return;
  if (HeadTask < 1) {RangeCentreTracker = Range; return;}

  HeadTaskRun = true; // prevent stack overflow due to nesting
  if (!servoHeadEn) {attachHeadServo();}
//  if (HeadTaskLast != HeadTask) {Serial.println("HeadTask " + String(HeadTask));}
  HeadTaskLast = HeadTask;
  switch(HeadTask) {
    // in this range of cases the head scans randomly left to right looking
    // if a target is detected by HeadMode code then do different actions
    case 1:
      // set random first head movement and head mode
      // also called when target has been lost
      RangeCentre = LtofMax;        // reset minimum range tracker
      RangeCentreTracker = LtofMax; // reset minimum range tracker
      RangeMinVal = LtofMax;        // reset minimum range tracker
      autoHeadOFF = -1;
      HeadAdj = false; // disable calculations for 1st half sweep
      if (MainTask < 4) {
        // target scanning & tracking mode
        HeadDir = random(2);  // set random head direction to start with
        HeadCyc = 4;          // if >0 then perform full left/right sweep
        HeadMode = 0;         // waits for target detection
        HeadTask++;           // next task is 2
      } else {
        // MainTask == 4 autonomous scanning mode
        TrgtDet = false;      // clear target detected flag
        CalcCnt = 0;          // counter used in autonomous mode
        HeadMode = 3;         // autonomous mode, looks for target
        HeadTask = 30;
      } break;
    case 2:
      // toggle head direction and set new target angle
      // if HeadCyc  > 0 perform full left/right sweeps
      // if HeadCyc == 0 perform sweeps to random angles in opposite quadrants
      if (HeadDir > 0) {
        // look to the right quadrant
        HeadDir = -1; // flag indicates moving to the right
        // set random head target angle between 90-160° in 10° steps
        if (HeadCyc > 0) {HeadTgt = Head_160;}
        else {HeadTgt = (((Head_160 - Head_90) * random(8))/7) + Head_90;}
      } else {
        // look to the left quadrant
        HeadDir = 1;  // flag indicates moving to the left
        // set random head target angle between 20-90° in 10° steps
        if (HeadCyc > 0) {HeadTgt = Head_20;}
        else {HeadTgt =  Head_90 - (((Head_90 - Head_20) * random(8))/7);}
      } 
      if (HeadCyc > 0) {HeadCyc--;} // if full-cycling reduce counter
      // set head steps based on angle to move through
      HeadTgtCnt = abs(((HeadTgt - HeadVal) * 50)/(Head_160 - Head_20));
      HeadTask++; break;
    case 3:
      // wait for head to reach end of movement
      if (HeadTgtCnt < 1) {
        // end of movement reached, set a random delay and HeadTask = 4
        HeadCnt = 10 + random(20); HeadTask++;
        if (HeadCyc > 0) {HeadCnt = 10;}  // full sweeps, so don't pause for long
      } break;
    case 4:
      // pause temporarily at this angle whilst HeadCnt > 0
      HeadCnt--; if (HeadCnt < 1) {
        if (HeadCyc > 0) {HeadTask = 2;}  // full sweep so go back to changing direction
        else {
          // complete the sweep
          if (HeadDir > 0) {
            // complete the move to the left
            HeadTgt = Head_20;
          } else {
            // complete the move to the right
            HeadTgt = Head_160;
          }
          HeadTgtCnt = abs(((HeadTgt - HeadVal) * 50)/(Head_160 - Head_20));
          HeadTask++;
        }
      } break;
    case 5:
      // wait for head to reach end of movement
      if (HeadTgtCnt < 1) {
        // end of movement reached, set a random delay and HeadTask = 4
        HeadCnt = 10 + random(20); HeadTask++;
        if (HeadCyc > 0) {HeadCnt = 10;}  // full sweeps, so don't pause for long
      } break;
    case 6:
      // pause temporarily at this angle whilst HeadCnt > 0
      HeadCnt--; if (HeadCnt < 1) {HeadTask = 2;} // change direction
      break;

    case 10:
      // scan a detected target whilst it remains in range, centred on the detected
      // angle as the starting point. Immediately stop the random sweep. Set the
      // sweep range to max.
      HeadMode = 1;           // set HeadMode to record minimum value
      HeadCnt = 0;            // clear the target lost counter
      HeadTgtCnt = 0;         // stop the current head movement
      RangeAng = RangeMinAng; // record head angle at which a target was detected
      RangeMinVal = LtofMax;  // reset minimum range tracker
      AngDif = AngDifMin;     // set the sweep angle delta value
      HeadAdj = false;        // prevent target angle adjustment until a full sweep is done
      HeadErrTrk = 0;         // zero error tracker
      HeadErrSt = false;      // clear the head error status flag
      HeadTask++; break;
    case 11:
      // this is a centre scan point, prior to sweeping right
      RangeCentre = min(int(Range),int(RangeMinVal)); // use the closest range
      RangePnt = true; // flag used by tracker function
      HeadTgt = min(Head_160, RangeAng + AngDif); // define sweep to right range
      SetHeadTgtCnt(HeadTgt - HeadVal); // set sweep speed
      HeadTask++; break;
    case 12:
      // turning to the right
      // if a target has been detected then slips away, we stop turning right,
      // and shift the target angle to the left, and don't calcuylate the error
      if (RangeMinVal <= LtofDet) {
        if (Range > (RangeMinVal + 10)) {HeadTgtCnt = 0; RangeAng -= AngDif;}
      }
      // wait for the rightmost target angle to be reached
      if (HeadTgtCnt < 1) {
        // rightmost point has been reached
//        if ((HeadCnt > 0) && (RangeMinVal < LtofMax)) {RangeAng = RangeMinAng;}
        if (HeadAdj) {HeadMode = 2;} // adjust only if full sweep has been completed
        HeadAdj = true; // a full sweep has now been completed
        HeadTask++; // move on once target is reached
      } RangeIntR += long(RangeMax - Range); // integrate the right-hand range values
      break;
    case 13:
      // adjustment may have moved the rightmost point, so move there before
      // starting a new sweep
      HeadTgt = min(Head_160, RangeAng + AngDif); // define sweep to right range
      SetHeadTgtCnt(HeadTgt - HeadVal); // set sweep speed
      HeadTask++; break;
    case 14:
      // wait for new rightmost end point to be reached  
      if (HeadTgtCnt < 1) {HeadTask++;} // rightmost point reached
      break;
    case 15:
      // now turn left towards the centre point
      RangePnt = true;                  // flag used by tracker function
      HeadTgt = RangeAng;               // centre angle is min range point
      SetHeadTgtCnt(HeadTgt - HeadVal); // set sweep speed
      RangeIntR = 0; RangeIntL = 0;     // clear the integrators
      RangeMinVal = LtofMax;            // reset minimum range tracker
      HeadTask++; break;
    case 16:
      // wait until the centre point has been reached
      // integrate the right-hand range value as we go
      if (HeadTgtCnt < 1) {HeadTask++;} // centre point achieved
      RangeIntR += long(RangeMax - Range); // integrate the right-hand range values
      break;
    case 17:
      // this is a centre scan point, prior to sweeping left
      RangeCentre = min(int(Range),int(RangeMinVal)); // use the closest range
      RangePnt = true; // flag used by tracker function
      HeadTgt = max(Head_20, RangeAng - AngDif);  // define sweep to left range
      SetHeadTgtCnt(HeadTgt - HeadVal);           // set sweep speed
      HeadTask++; break;
    case 18:
      // turning to the right
      // if a target has been detected then slips away, we stop turning left,
      // and shift the target angle to the right, and don't calculate the error
      if (RangeMinVal <= LtofDet) {
        if (Range > (RangeMinVal + 10)) {HeadTgtCnt = 0; RangeAng += AngDif;}
      }
      // wait for the leftmost target angle to be reached
      // integrate the left-hand range value as we go
      if (HeadTgtCnt < 1) {
        // leftmost point has been reached
//        if ((HeadCnt > 0) && (RangeMinVal < LtofMax)) {RangeAng = RangeMinAng;}
        if (HeadAdj) {HeadMode = 2;}  // calculate centre adjustments based on this full sweep
        HeadAdj = true; // a full sweep has now been completed
        HeadTask++;     // move on once target is reached
      } RangeIntL += long(RangeMax - Range); // integrate the left-hand range values
      break;
    case 19:
      // adjustment may have moved the leftmost point, so move there before
      // starting a new sweep
      HeadTgt = max(Head_20, RangeAng - AngDif);  // define sweep to right range
      SetHeadTgtCnt(HeadTgt - HeadVal);           // set sweep speed
      HeadTask++; break;
    case 20:
      // wait for new leftmost end point to be reached  
      if (HeadTgtCnt < 1) {HeadTask++;} // rightmost point reached
      break;
    case 21:
      // now turn right towards the centre point
      RangePnt = true; // flag used by tracker function
      HeadTgt = RangeAng;               // centre angle on min range point
      SetHeadTgtCnt(HeadTgt - HeadVal); // set sweep speed
      RangeIntR = 0; RangeIntL = 0;     // clear the integrators
      RangeMinVal = LtofMax;            // reset minimum range tracker
      HeadTask++; break;
    case 22:
      // wait until the centre point has been reached
      // integrate the left-hand range value as we go
      if (HeadTgtCnt < 1) {HeadTask = 11;}  // centre point achieved
      RangeIntL += long(RangeMax - Range);             // integrate the left-hand range values
      break;

    case 30:
      // these set of cases perform simple left to right scanning for autonomous
      // target detection.
      // centre point continue towards left-most point
      if (HeadAdj) {CalcHeadAng();} // calculate adjustments based on this half sweep
      RangeIntL = 0; // clear left-hand integrator
      HeadTgt = Head_20; HeadTgtCnt = HeadSpd; HeadTask++; break;
    case 31:
      // wait for new leftmost end point to be reached  
      if (HeadTgtCnt < 1) {HeadTask++;} // left-most point reached
      RangeIntL += long(Range); // integrate the left-hand range values
      break;
    case 32:
      // left-most point reached
      // turn right towards centre position
      CheckMinRange();
      HeadTgt = Head_90; HeadTgtCnt = HeadSpd; HeadTask++; break;
    case 33:
      // wait for centre point to be reached  
      if (HeadTgtCnt < 1) {HeadTask++;} // centre point reached
      RangeIntL += long(Range); // integrate the left-hand range values
      break;
    case 34:
      // centre point continue right towards right-most position
      CalcHeadAng(); // calculate adjustments based on this half sweep
      RangeIntR = 0; // clear right-hand integrator
      HeadTgt = Head_160; HeadTgtCnt = HeadSpd; HeadTask++; break;
    case 35:
      // wait for right-most point to be reached  
      if (HeadTgtCnt < 1) {HeadTask++;} // right-most point reached
      RangeIntR += long(Range); // integrate the right-hand range values
      break;
    case 36:
      // right-most point reached
      // turn towards centre position
      CheckMinRange();
      HeadAdj = true; // full half sweeps can now be integrated
      HeadTgt = Head_90; HeadTgtCnt = HeadSpd; HeadTask++; break;
    case 37:
      // wait for centre point to be reached  
      if (HeadTgtCnt < 1) {HeadTask = 30;} // right-most point reached
      RangeIntR += long(Range); // integrate the right-hand range values
      break;
  }

  switch(HeadMode) {
    // this section oversees what results from above HeadTask scanning code
    // here we use Range values to switch between different head scanning tasks
    case 0:
      // wait here whilst no target is being detected whilst scanning
      RangeCentre = Range;
      // track the minimum whilst sweeping
      if (Range < RangeMinVal) {RangeMinVal = Range; RangeMinAng = HeadVal;}
      if (RangeMinVal <= LtofDet) {
        // a target has been detected, so wait for a minimum
        // once range is increasing, switch to close range scanning
        // we include a 10mm dead band for noise
        if (Range > (RangeMinVal + 10)) {HeadTask = 10;}
      }
      break;
    case 1:
      // we are performing a target scan
      // only record minimum during full sweep
      if (Range < RangeMinVal) {RangeMinVal = Range; RangeMinAng = HeadVal;}
      break;
    case 2:
      // recalculate centre angle at both sweep end points
      // the adjustment is based on the difference in L/R integrators
      if ((RangeIntR + RangeIntL) != 0) {
        // integrators have accumulated values
        HeadErr = ((RangeIntR - RangeIntL) * 100)/(RangeIntR + RangeIntL);
        RangeAng = RangeAng + int(HeadErr);
        if (abs(HeadErr) > HeadErrTrk) {HeadErrTrk++;} else {HeadErrTrk--;}
      }
      HeadErrTrip = abs(abs(HeadErr) - HeadErrTrk);
//      Serial.print(RangeIntL); Serial.print("\t"); Serial.println(RangeIntR);
//      Serial.print(HeadErrTrip); Serial.print("\t"); Serial.println(AngDif);
      // adjust sweep width depending on angle adjustment needed
      if ((HeadErrTrip > 10) && (HeadErrSt)) {
        // significant head angle adjustment so increase sweep angle
        AngDif = AngDifMax; // increase the sweep angle if target is moving
        HeadErrSt = false;  // clear the head error status flag
      } else {
        // decrease the sweep angle progressively to reduce the error
        // then wait for  errors to occur due to target movement
        if (AngDif > AngDifMin) {
          AngDif = max(AngDifMin,AngDif - (AngDif/8));
        } else if (HeadErrTrip < 4) {HeadErrSt = true;}
      }
      // check for lost target in either phase
      if (RangeMinVal >= LtofDet) {
        // nothing detected in this half cycle, so increment a counter
        HeadCnt++;  // increment the target lost counter
        // could be slightly off centre so widen the sweep
        if (HeadCnt == 1) {AngDif = AngDifMax + 300;};  // no target detected in 1/2 cycle
        if (HeadCnt >= 2) {HeadTask = 1;}               // 1 dead cycles, target has been lost
      } else {HeadCnt = 0;}                             // zero the target lost counter
      RangeAng = min(int(RangeAng),Head_160); RangeAng = max(int(RangeAng),Head_20);
      HeadMode = 1; break;
    
    case 3:
      // mode used for autonomous
      if (Range < RangeMinVal) {
        RangeMinVal = Range;
        // check to see if something has been detected
        if (RangeMinVal <= LtofDet) {TrgtDet = true;}
      } break;
  }
  // before leaving track the centre positions
  if (RangePnt) {
    RangePnt = false; // reset the point flag
    anyInt = (RangeCentre - RangeCentreTracker)/2; RangeCentreTracker += anyInt;
//  RangeCentreTracker = RangeCentreTracker + ((RangeCentre - RangeCentreTracker)/8);
  }
  // release stack overflow block and allow this function to be called again from
  // the main loop()
  HeadTaskRun = false;
}

// --------------------------------------------------------------------------------

void GetHeadAngle() {
  // returns HeadVal converted to an angle as HeadAng
  if (HeadVal < Head_90) {HeadAng = map(HeadVal,Head_20,Head_90,20,90);}
  else {HeadAng = map(HeadVal,Head_90,Head_160,90,160);}
}

// --------------------------------------------------------------------------------

void GetSteeringDrive(int zDir) {
  // calculates steering angle when head is active
  // zDir defines direction, 1 == forward, -1 == backwards
  // at centre RangeAng == Head_0, full left == Head_20, full right == Head_160
  WalkLftDrive = 128; WalkRgtDrive = 128; // set drive factors to max as default
  if (!HeadActive) {return;} // head is not active so no steering needed
  if (zDir == 0) {return;} // set drives equal
  if (zDir > 0) {
    // walking forwards
    if (RangeAng >= Head_90) {
      // target is on the right-hand side
      WalkRgtDrive = map(RangeAng,Head_160,Head_90,0,128);
    } else {
      // target is on the left-hand side
      WalkLftDrive = map(RangeAng,Head_20,Head_90,0,128);
    }
  } else {
    // walking backwards
    if (RangeAng >= Head_90) {
      // target is on the right-hand side
      WalkLftDrive = map(RangeAng,Head_160,Head_90,0,128);
    } else {
      // target is on the left-hand side
      WalkRgtDrive = map(RangeAng,Head_20,Head_90,0,128);
    }
  }
  WalkLftDrive = max(WalkLftDrive,0); WalkLftDrive = min(WalkLftDrive,128);
  WalkRgtDrive = max(WalkRgtDrive,0); WalkRgtDrive = min(WalkRgtDrive,128);
}

// --------------------------------------------------------------------------------

void GoToAngles(int16_t zA0,int16_t zA1,int16_t zA2,int16_t zA3,int16_t zA4,int16_t zA5,int16_t zA6,int16_t zA7,int16_t zA8,int16_t zA9,int16_t zAA,int16_t zAB,int16_t zCnt) {
  // move to specific servo angles, or -1 angle means no change
  // all 12 angles in degrees are converted to servo PWM target values
  // number of 20ms steps in movement set by zCnt
  // if zCnt > 0 then perform the moves immediately from here
  // if zCnt < 0 then moves are controlled elswewhere
  if (ESC) return; // exit if this flag is set
  
  if (!ServoEn) {attachServos(0);}
  if (zA0 >= 0) {
    // Servo 0 front right claw
    if (zA0 >= 90) {servoTgt[ 0] = map(zA0,90,223,Ang0_90,Ang0_223);}
    else {servoTgt[ 0] = map(zA0,69,90,Ang0_69,Ang0_90);}
  }
  if (zA1 >= 0) {
    // Servo 1 front right shoulder
    if (zA1 >= 90) {servoTgt[ 1] = map(zA1,90,120,Ang1_90,Ang1_120);}
    else {servoTgt[ 1] = map(zA1,21,90,Ang1_21,Ang1_90);}
  }
  if (zA2 >= 0) {
      // Servo 2 centre right claw
    if (zA2 >= 90) {servoTgt[ 2] = map(zA2,90,223,Ang2_90,Ang2_223);}
    else {servoTgt[ 2] = map(zA2,69,90,Ang2_69,Ang2_90);}
  }
  if (zA3 >= 0) {
    // Servo 3 centre right shoulder
    if (zA3 >= 90) {servoTgt[ 3] = map(zA3,90,150,Ang3_90,Ang3_150);}
    else {servoTgt[ 3] = map(zA3,30,90,Ang3_30,Ang3_90);}
  }
  if (zA4 >= 0) {
    // Servo 4 rear right shoulder
    if (zA4 >= 90) {servoTgt[ 4] = map(zA4,90,180,Ang4_90,Ang4_180);}
    else {servoTgt[ 4] = map(zA4,68,90,Ang4_68,Ang4_90);}
  }
  if (zA5 >= 0) {
    // Servo 5 rear right claw
    if (zA5 >= 90) {servoTgt[ 5] = map(zA5,90,223,Ang5_90,Ang5_223);}
    else {servoTgt[ 5] = map(zA5,69,90,Ang5_69,Ang5_90);}
  }
  
  if (zA6 >= 0) {
    // Servo 6 rear left shoulder
    if (zA6 >= 90) {servoTgt[ 6] = map(zA6,90,180,Ang6_90,Ang6_180);}
    else {servoTgt[ 6] = map(zA6,68,90,Ang6_68,Ang6_90);}
  }
  if (zA7 >= 0) {
    // Servo 7 rear left claw
    if (zA7 >= 90) {servoTgt[ 7] = map(zA7,90,223,Ang7_90,Ang7_223);}
    else {servoTgt[ 7] = map(zA7,69,90,Ang7_69,Ang7_90);}
  }
  if (zA8 >= 0) {
    // Servo 8 centre left claw
    if (zA8 >= 90) {servoTgt[ 8] = map(zA8,90,223,Ang8_90,Ang8_223);}
    else {servoTgt[ 8] = map(zA8,69,90,Ang8_69,Ang8_90);}
  }
  if (zA9 >= 0) {
    // Servo 9 centre left shoulder
    if (zA9 >= 90) {servoTgt[ 9] = map(zA9,90,150,Ang9_90,Ang9_150);}
    else {servoTgt[ 9] = map(zA9,30,90,Ang9_30,Ang9_90);}
  }
  if (zAA >= 0) {
    // Servo 10 front left claw
    if (zAA >= 90) {servoTgt[10] = map(zAA,90,223,AngA_90,AngA_223);}
    else {servoTgt[10] = map(zAA,69,90,AngA_69,AngA_90);}
  }
  if (zAB >= 0) {
    // Servo 11 front left shoulder
    if (zAB >= 90) {servoTgt[11] = map(zAB,90,120,AngB_90,AngB_120);}
    else {servoTgt[11] = map(zAB,21,90,AngB_21,AngB_90);}
  }
  checkTgts(); // ensure target values are within limits
  TgtCnt = abs(zCnt);
  if (zCnt > 0) {loopWhileTgtCnt();}  // included here for code efficiency
//  Serial.println("GoToAngles Set");
//  Serial.println(TgtCnt);
}

// --------------------------------------------------------------------------------

void GoToReady(int zCnt) {
  // move the legs to the 90 resting position
  // number of 20ms steps in movement set by zCnt
  if (ESC) return;
//  Serial.println("> GoToRest()");
// servo    (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type     ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
  GoToAngles(   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,   90,-zCnt); // go to the resting position
  autoRest = autoRestMax; // no need to rest from this position
  SetLedMode(2);
//  DetCnt = 50;            // detach leg servos 1 sec after movement has stopped
}

// --------------------------------------------------------------------------------

void GoToStand(int zCnt) {
  // move to the normal 60/120 standing position
  // number of 20ms steps in movement set by zCnt
  if (ESC) return;
  Serial.println(F("> GoToStand()"));
// servo    (      0,     1,      2,     3,     4,     5,     6,      7,      8,     9,     10,    11) servo reference
// type     (   Claw, Shldr,   Claw, Shldr, Shldr,  Claw, Shldr,   Claw,   Claw, Shldr,   Claw, Shldr)
  GoToAngles(     90,    60,     90,    90,   120,    90,   120,     90,     90,    90,     90,    60,   -zCnt); // go to the resting position
  SetLedMode(1);
  DetCnt = 100; // detach leg servos 2 sec after movement has stopped
}

// --------------------------------------------------------------------------------

void GoToStore(int zCnt) {
  // move the legs to the powered off stored position, head centred
  // number of 20ms steps in movement set by zCnt
  if (ESC) return;

  SetMainTask(0); WiFiEn = false; HexBotActive = false;
  HeadTask = 0; HeadActive = false; HeadToRest(); HeadMoveNow();  // centre head servo
//  Serial.println("> GoToRest()");
// servo    (      0,     1,      2,     3,     4,      5,     6,      7,      8,     9,     10,    11) servo reference
// type     (   Claw, Shldr,   Claw, Shldr, Shldr,   Claw, Shldr,   Claw,   Claw, Shldr,   Claw, Shldr)
  GoToAngles(ClawRst,   110,ClawRst,   150,   150,ClawRst,   150,ClawRst,ClawRst,   150,ClawRst,   110,-zCnt); // go to the resting position
  MoveNow();
  autoRest = autoRestMax;       // no need to auto-rest from this position
  DetCnt = 50;                  // detach leg servos 1 sec after movement has stopped
}

// --------------------------------------------------------------------------------

void HeadMoveEngine() {
  // called nominally every 20ms from the main loop to perform head movement tasks
  if (HeadPause > 0) {
    // reduce headPause to zero before doing anything
    // headPause counts in 20ms increments, 50 == 1 second
    HeadPause--;
  } else {
    // we aren't pausing, so do we need to move the head
    if (HeadTgtCnt > 0) {
      // yes a head move is in progress
//      Serial.println(headTgtCnt);
      if (autoHeadOFF >= 0) {autoHeadOFF = 20;} // set auto-OFF timeout
      atHeadTgt = true; // set a test flag for achieving target values
      if (HeadTgtCnt == 1) {
        // move head immediately to the target value
        HeadVal = HeadTgt; // set values to target values
        servoHead.writeMicroseconds(HeadVal);
      } else {
        // move head value progressively towards the target value
        servoDelta = HeadTgt - HeadVal;
        if (abs(servoDelta) > 0) {
          HeadVal = HeadVal + (servoDelta / HeadTgtCnt);
          servoHead.writeMicroseconds(HeadVal); atHeadTgt = false;
        } if (atHeadTgt) {HeadTgtCnt = 0;} // already at target value so clear the move count
      }
      HeadTgtCnt--; // progressivley reduce the target factor
//      Serial.println(HeadVal);
//      Serial.println(String(HeadVal) + "," + String(Range) + "," + String(RangeAng) + "," + String(HeadErr));
    }
  }
}

// --------------------------------------------------------------------------------

void HeadMoveNow() {
  // called to move head to target position, ignoring all other inputs
  // returns to calling function once TgtCnt = 0
  HeadPause = 0;
  while (HeadTgtCnt > 0) {HeadMoveEngine(); delay(20);}
  synchLoopTimers(0); // re-synch loop timers
}

// --------------------------------------------------------------------------------

void HeadToAngle(int zA, int zCnt) {
  // move to specific servo angle
  // number of steps in head movement set by zCnt
  if (ESC) return; // exit if this flag is set
  
  if (!servoHeadEn) {attachHeadServo();}
  // assume non-linearity and map angle into the correct quadrant for best
  // accuracy
  zA = max(zA,20); zA = min(zA,160);  // constrain to limits 
  if (zA >= 90) {
    HeadTgt = map(zA,90,160,Head_90,Head_160);
  } else {
    HeadTgt = map(zA,20,90,Head_20,Head_90);
  }
  HeadTgtCnt = zCnt;
//  Serial.println("HeadToAngles Set" + String(HeadTgtCnt));
}

// --------------------------------------------------------------------------------

void HeadToRest() {
  // move the head to the straight forward rest position
  HeadToAngle(90,10); autoHeadOFF = 5; // centre head servo, then power off
}

// --------------------------------------------------------------------------------

void JoyMoveBwd() {
  // called from main loop to respond to joystick demands
  // this function walks and steers in a backwards direction
  WalkLftDrive = 128; WalkRgtDrive = 128;
//  Serial.println("JoyMoveBwd");
  if (Walk == 0) {
    // the start of a new walk movement
    DispMode = 3; WalkFreq = 0;
    if (WalkLast == 1) {
      // previously moving forward, so reverse counter and go backwards
      SetLegsPrevious(); Walk = 3;
    } else if (WalkLast != 3) {
      // previously walking in a different direction
//      Serial.println(F("Walking Backwards"));
      walkInterval = 30000; StartWalk(3); SetLedMode(23);
    } else {
      // continue walking in same direction
      SetLegsPrevious(); Walk = WalkLast;
    } JoyYi = 0; JoyYT = true; SetLedMode(23);
  } else {
    // ensure joystick values are in the required range
    if (JoyX < 127) {JoyXV = 128 - JoyX;} // convert -ve vector to +ve
    else {JoyXV = JoyX - 127;} // ensure JoyX = 0 - 128
    if (JoyY < 128) {JoyYV = 128 - JoyY;} // convert -ve vector to +ve
    else {JoyYV = JoyY - 127;} // ensure JoyY = 0 - 128
    JoySpd = JoyYV; if (JoyXV > JoyYV) {JoySpd = JoyXV;}
    // we use the larger of the two joystick values to set the speed
    // JoyYi is a tracking filter to slug the initial joystick response
    if (JoyYT) {
      // initially slug the response until it catches up
      if (JoyYi < (JoySpd-2)) {JoyYi += 2;}
      else if (JoyYi > (JoySpd+2)) {JoyYi -= 2;}
      else {JoyYi = JoySpd; JoyYT = false;}
    } else {JoyYi = JoySpd;}
    WalkFreq = map((unsigned long)JoyYi,0,128,1,WalkMax);
    walkInterval = 240000/WalkFreq;
    // now apply steering
    if (JoyX <= 127) {WalkLftDrive = JoyX + 1;}
    else {WalkRgtDrive = 128-(JoyX-128);}
    if ((WalkSpeed == 1) && (C_Dn == 25)) {
      // 'C' held during walking
      LegUp = LegUpMax; WalkLftDrive /= 2; WalkRgtDrive /= 2;
    } else {
      LegUp = LegUpMin;
    }
  }
}

// --------------------------------------------------------------------------------

void JoyMoveFwd() {
  // called from main loop to respond to joystick demands
  // this function walks and steers in a backwards direction
  WalkLftDrive = 128; WalkRgtDrive = 128;
//  Serial.println("JoyMoveFwd");
  if (Walk == 0) {
    // the start of a new walk movement
    DispMode = 2; WalkFreq = 0;
    if (WalkLast == 3) {
      // previously moving backward, so reverse counter and go backwards
      SetLegsPrevious(); Walk = 1;
    } else if (WalkLast != 1) {
      // previously walking in a different direction
//      Serial.println(F("Walking Forwards"));
      walkInterval = 30000; StartWalk(1); SetLedMode(21);
    } else {
      // continue walking in same direction
      SetLegsPrevious(); Walk = WalkLast;
    } JoyYi = 0; JoyYT = true; SetLedMode(21);
  } else {
    // ensure joystick values are in the required range
    if (JoyX < 127) {JoyXV = 128 - JoyX;} // convert -ve vector to +ve
    else {JoyXV = JoyX - 127;} // ensure JoyX = 0 - 128
    if (JoyY < 128) {JoyYV = 128 - JoyY;} // convert -ve vector to +ve
    else {JoyYV = JoyY - 127;} // ensure JoyY = 0 - 128
    JoySpd = JoyYV; if (JoyXV > JoyYV) {JoySpd = JoyXV;}
    // we use the larger of the two joystick values to set the speed
    // JoyYi is a tracking filter to slug the initial joystick response
    if (JoyYT) {
      // initially slug the response until it catches up
      if (JoyYi < (JoySpd-2)) {JoyYi += 2;}
      else if (JoyYi > (JoySpd+2)) {JoyYi -= 2;}
      else {JoyYi = JoySpd; JoyYT = false;}
    } else {JoyYi = JoySpd;}
    WalkFreq = map((unsigned long)JoyYi,0,128,1,WalkMax);
    walkInterval = 240000/WalkFreq;
    // now apply steering
    if (JoyX <= 127) {WalkLftDrive = JoyX;}
    else {WalkRgtDrive = 128-(JoyX-128);}
//    Serial.println(String(WalkLftDrive) + "\t" + String(WalkRgtDrive));
    // now set leg up height if in lowest speed
    if ((WalkSpeed == 1) && (C_Dn == 25)) {
      // 'C' held during walking
      LegUp = LegUpMax; WalkLftDrive /= 2; WalkRgtDrive /= 2;
    } else {
      LegUp = LegUpMin;
    }
  }
}

// --------------------------------------------------------------------------------

void JoyMoveLft() {
  // called from main loop to respond to joystick demands
  // this function walks and steers in a leftwards direction
  WalkLftDrive = 128; WalkRgtDrive = 128;
  Serial.println("JoyMoveLft");
}

// --------------------------------------------------------------------------------

void JoyMoveRgt() {
  // called from main loop to respond to joystick demands
  // this function walks and steers in a rightwards direction
  WalkLftDrive = 128; WalkRgtDrive = 128;
  Serial.println("JoyMoveRgt");
}

// --------------------------------------------------------------------------------

void JoyTurnLft() {
  // called from main loop to respond to joystick demands
  // this function neutral turns in a leftwards direction
  Serial.println("JoyTurnLft");
  if (Walk == 0) {
    DispMode = 5; WalkFreq = 0;
    // the start of a new walk movement
    if (WalkLast != 5) {
      // previously walking in a different direction
//      Serial.println(F("Turning Left"));
      walkInterval = 60000; StartWalk(5); SetLedMode(25);
    } else {
      // continue walking in same direction
      SetLegsPrevious(); Walk = WalkLast;
    } JoyYi = 0; JoyYT = true; SetLedMode(25);
  } else {
    // ensure joystick values are in the required range
    if (JoyX < 127) {JoyXV = 128 - JoyX;} // convert -ve vector to +ve
    else {JoyXV = JoyX - 127;} // ensure JoyXV = 0 - 128
    // we use the joystick value to set the speed
    JoySpd = JoyXV;
    // JoyYi is a tracking filter to slug the initial joystick response
    if (JoyYT) {
      // initially slug the response until it catches up
      if (JoyYi < (JoySpd-2)) {JoyYi += 2;}
      else if (JoyYi > (JoySpd+2)) {JoyYi -= 2;}
      else {JoyYi = JoySpd; JoyYT = false;}
    } else {JoyYi = JoySpd;}
    WalkFreq = map((unsigned long)JoyYi,0,128,1,WalkMax);
    walkInterval = 240000/WalkFreq;
  }
}

// --------------------------------------------------------------------------------

void JoyTurnRgt() {
  // called from main loop to respond to joystick demands
  // this function turns in a rightwards direction
  Serial.println("JoyTurnRgt");
  if (Walk == 0) {
    DispMode = 6; WalkFreq = 0;
    // the start of a new walk movement
    if (WalkLast != 6) {
      // previously walking in a different direction
//      Serial.println(F("Turning Right"));
      walkInterval = 60000; StartWalk(6); SetLedMode(26);
    } else {
      // continue walking in same direction
      SetLegsPrevious(); Walk = WalkLast;
    } JoyYi = 0; JoyYT = true; SetLedMode(26);
  } else {
    // ensure joystick values are in the required range
    if (JoyX < 127) {JoyXV = 128 - JoyX;} // convert -ve vector to +ve
    else {JoyXV = JoyX - 127;} // ensure JoyXV = 0 - 128
    // we use the joystick value to set the speed
    JoySpd = JoyXV;
    // JoyYi is a tracking filter to slug the initial joystick response
    if (JoyYT) {
      // initially slug the response until it catches up
      if (JoyYi < (JoySpd-2)) {JoyYi += 2;}
      else if (JoyYi > (JoySpd+2)) {JoyYi -= 2;}
      else {JoyYi = JoySpd; JoyYT = false;}
    } else {JoyYi = JoySpd;}
    WalkFreq = map((unsigned long)JoyYi,0,128,1,WalkMax);
    walkInterval = 240000/WalkFreq;
  }
}

// --------------------------------------------------------------------------------

void Leg_Engine() {
  //###############################################################################
  //
  //  Leg Engine tasks
  //
  //###############################################################################
  // called from the main loop if not walking and TgtCnt > 0
  if (autoOFF >= 0) {autoOFF = 5;} // set auto-OFF timeout
  if (TgtCnt == 1) {
    // move values immediately to the target value
    for (int zP = 0; zP < 12;zP++) {
      servoVal[zP] = servoTgt[zP]; // set values to target values
      servoInst[zP].writeMicroseconds(servoVal[zP]);
    }
  } else {
    // move values progressively towards the target values
    atTgt = true; // set a test flag for achieving target values
    for (int zP = 0; zP < 12;zP++) {
      servoDelta = servoTgt[zP] - servoVal[zP];
      if (abs(servoDelta) > 0) {
        servoVal[zP] = servoVal[zP] + (servoDelta / TgtCnt);
        servoInst[zP].writeMicroseconds(servoVal[zP]); atTgt = false;
      }
    } if (atTgt) {TgtCnt = 0;} // already at target values so clear the move count
  }
  if (TgtCnt > 0) {TgtCnt--;} // progressivley reduce the target factor
//    Serial.println(tgtCnt);
}

// --------------------------------------------------------------------------------

void MainTask_Autonomous() {
  // autonomous discovery mode, stays in this mode until reset
  // starts with looking in front of itself
  // if clear, then move forward, a random distance if fixed head
  // if not clear, rotate and look again
  // if rotating head, walk and scan at the same time
  if (MainRun) {return;}
  MainRun = true; // prevent re-entry of task
  
//  Serial.println("SubTask=" + String(SubTask)); 
  switch (SubTask) {
    case 0: // initialise
      Serial.println("> MainTask_Autonomous()"); 
      DispMode = 4; AutoMode = 0; automRnd = 1; SubTask++; break;
    case 1:
      // set legs for going forward, prior to radar scanning
      automTurn = 0; Walk = 0; walkInterval = 18000; StartWalk(-1); 
      if (HeadActive) {
        SubTask = 50; // autonomous with rotating head
      } else {
        SubTask = 2; // autonomous with fixed head, twisting body
      } SetLedMode(5);
      break;
    case 2:
      //###############################################################################
      //
      //  Fixed head twisting body
      //
      //###############################################################################
      // twist & scan randomly and turning code
      // Start with ranges at max.
      Range = LtofMax;
      RangeCentre = LtofMax;  RangeLeft = LtofMax; RangeRight = LtofMax;
      SetLedMode(5); walkInterval = 18000; AutoMode = 2;
      if (random(2) < 1) {
        // twist left, then right, taking radar readings
        // twist to the left by 20°
//        Serial.println("Looking left..."); 
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
        GoToAngles(   -1,  110,   -1,  110,  110,   -1,   70,   -1,   -1,   70,   -1,   70,60);
        loopWhileTgtCnt(); delayLoop(10);
        if (Range <= LtofDet) {RangeLeft = Range;}
//        Serial.println(Range); 
        // twist to the right by 20°
//        Serial.println("Looking right..."); 
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
        GoToAngles(   -1,   70,   -1,   70,   70,   -1,  110,   -1,   -1,  110,   -1,  110,30);
        loopWhileTgtCnt(); delayLoop(10);
        if (Range <= LtofDet) {RangeRight = Range;}
//        Serial.println(Range); 
      } else {
        // twist right, then left, taking radar readings
        // twist to the right by 20°
//        Serial.println("Looking right..."); 
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
        GoToAngles(   -1,   70,   -1,   70,   70,   -1,  110,   -1,   -1,  110,   -1,  110,30);
        loopWhileTgtCnt(); delayLoop(10);
        if (Range <= LtofDet) {RangeRight = Range;}
//        Serial.println(Range); 
        // twist to the left by 20°
//        Serial.println("Looking left..."); 
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
        GoToAngles(   -1,  110,   -1,  110,  110,   -1,   70,   -1,   -1,   70,   -1,   70,60);
        loopWhileTgtCnt(); delayLoop(10);
        if (Range <= LtofDet) {RangeLeft = Range;}
//        Serial.println(Range); 
      }
      // back to centre
//      Serial.println(F("Looking forward...")); 
// servo        (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type         ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
      GoToAngles(   -1,   90,   -1,   90,   90,   -1,   90,   -1,   -1,   90,   -1,   90,30);
      loopWhileTgtCnt(); delayLoop(10);
      if (Range <= LtofDet) {RangeCentre = Range;}
//      Serial.println(Range); 
      // now check range readings and respond
      if (automTurn == 0) {
        // haven't just turned, so determine direction
        if (RangeLeft <= 325) {
          // something is to my left so turn right and repeat process measurements
          automRnd = 1; // reset walking to short distance at first
          if (RangeRight > RangeLeft) {
            // object on the left is closer
  //          Serial.println("Turning Right..."); 
            AutoMode = 3;
            NeutralTurnRight(60,15000); automTurn = 1; Walk = 0;
          } else {
            // object on the right is closer
  //          Serial.println("Turning Left..."); 
            AutoMode = 3;
            NeutralTurnLeft(60,15000); automTurn = -1; Walk = 0;
          } break;
        }
        else if (RangeRight <= 325) {
          // something is to my right so turn left and repeat process measurements
          automRnd = 1; // reset walking to short distance at first
          if (RangeLeft > RangeRight) {
            // object on the right is closer
            // TurnLeft(zAng,zStep1,zSpeed)
  //          Serial.println("Turning Left..."); 
            AutoMode = 3;
            NeutralTurnLeft(60,15000); automTurn = -1; Walk = 0;
          } else {
            // object on the left is closer
            // TurnRight(zAng,zStep1,zSpeed)
  //          Serial.println("Turning Right..."); 
            AutoMode = 3;
            NeutralTurnRight(60,15000); automTurn = 1; Walk = 0;
          } break;
        }
        else if (RangeCentre <= 325) {
          // something is in front of me, but sides are clear
          // so turn randomly left/right
  //        Serial.println(F("Object ahead...")); 
          automRnd = 1; // reset walking to short distance at first
          if (random(2) > 0) {
  //          Serial.println(F("Turning Left...")); 
            AutoMode = 3;
            NeutralTurnLeft(90,15000); automTurn = -1; Walk = 0;
          } else {
  //          Serial.println(F("Turning Right...")); 
            AutoMode = 3;
            NeutralTurnRight(90,15000); automTurn = 1; Walk = 0;
          }
        } 
        else {SubTask = 3; break;}
      }
      else if ((RangeCentre <= 325) || (RangeLeft <= 325) || (RangeRight <= 325)) {
        // despite turning an object has still been detected
        // so we continue turning in the same direction
        automRnd = 1; // reset walking to short distance at first
        if  (automTurn > 0) {
          // just turned to the right, so continue turning if object is present
          AutoMode = 3;
          NeutralTurnRight(60,15000); Walk = 0;
        } else {
          // just turned to the left, so continue turning if object is present
          AutoMode = 3;
          NeutralTurnLeft(60,15000); Walk = 0;
        }
      }
      else {SubTask = 3;}
      break;
    case 3:
      // all is clear so walk for a random distance, then re-measure
      // random distance increases provided nothing is detected along the way
//        Serial.println(F("Walking forward..."));
      walkInterval = 15000 + (1000 * random(7)); // vary walking speed
      StartWalk(1); SetLedMode(5); AutoMode = 1;
      loopWhileWalkEcho(64 * automRnd);
      if (automRnd < 10) {automRnd++;} // increase walking range
      delayLoop(200); // short pause before range scanning
      SubTask = 1; // restart scanning process
      break;


    case 50:
      //###############################################################################
      //
      //  Moving head moving body
      //
      //###############################################################################
      // autonomouse with moving head
      // at this point the robot is stationary and checking for a wall
      // the head is constantly scanning left to right
      if (CalcCnt > 3) {
        // head has performed at least 4 half-sweeps
        CalcCnt = 0;
        if ((TrgtDet) && (RangeMinMem <= LtofStop)) {
          // if wall or target is close go into 360 turning mode
          // whilst in this mode move the head more rapidly
          HeadInterval = 10000;
          if (RangeAng >= Head_90) {
            // wall is closer on the left
            AutoMode = 3;
             // turn 360 degrees right whilst a wall/target is in range
             // if the wall disappears an ESC is generated
            NeutralTurnRight(60,15000); // start with a short turn no range detection
            NeutralTurnRightEcho(360,15000);
            Walk = 0;
          } else {
            // wall is closer on the right
            AutoMode = 3;
             // turn 360 degrees left whilst a wall/target is in range
             // if the wall disappears an ESC is generated
            NeutralTurnLeft(60,15000); // start with a short turn no range detection
            NeutralTurnLeftEcho(360,15000);
            Walk = 0;
          }
          if (!ESC) {SubTask = 100;} // no way out found
          else {
            ESC = false; HeadInterval = 20000;
            WalkLftDrive = 128; WalkRgtDrive = 128;
            AutoMode = 4; StartWalk(1); SetLedMode(5); SubTask++;
          }
        } else {
          // no target or wall was detected so start walking
          WalkLftDrive = 128; WalkRgtDrive = 128;
          AutoMode = 4; StartWalk(1); SetLedMode(5); SubTask++;
        }
      } break;
    case 51:
      // walking whilst looking for targets/walls
      if (CalcCnt > 0) {
        // a new sweep has been completed so recalculate steering
        CalcCnt = 0;
        GetSteeringDrive(1); // get steering in forward direction
      }
      // whilst walking constrantly look for a close wall or target
      if (RangeMinVal <= LtofStop) {
        // a target or wall was detected close range so stop
        Walk = 0; SetLedMode(5);
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr)
        GoToAngles(LegDn,   -1,LegDn,   -1,   -1,LegDn,   -1,LegDn,LegDn,   -1,LegDn,   -1,-1); // put all feet down immediately
        SubTask = 50; // go back to start point
      }
      break;


    case 100:
      // here because turning 360 did not find a way out
      HeadTask = 0; HeadToRest();
      automCnt = 500 + random(1000); // create a delay for a minimum of 10 seconds
      break;
    case 101:
      // wait here then restart
      automCnt--; if (automCnt < 1) {SubTask = 0;}
      break;
  } MainRun = false;
}

// --------------------------------------------------------------------------------

void MainTask_LTOF() {
  // HexBot goes to stand to demonstrate LTOF ranging  
  GoToStand(50); MoveNow();
  SetLedMode(5); MainTask = 0;
}

// --------------------------------------------------------------------------------

void MainTask_TrackWall(int zMode) {
  // dual function depending on zMode value
  // zMode = 0 backs away from an approaching target
  // zMode = 1 maintains a fixed distance from a wall, stays in this mode until reset
  // refer to Calibration document for ranging diagrams
  if (MainRun) {return;}
  MainRun = true; // prevent re-entry of task
//  Serial.println("Walk= " + String(Walk) + "\tSubTask=" + String(SubTask));
  switch (SubTask) {
    case 0: // initialise, set range limits and LED mode
      // R0 - minimum range, stopping point
      // RM - max range, object detected
      R0 = 55; RM = 335; Walked = 0; SetLedMode(5);
      WalkLftDrive = 128; WalkRgtDrive = 128;
      if (zMode == 0) {
        Serial.println(F("> MainTask_BackAway()"));
        // R1 - range below which we slow down
        // R2 - range which detemines max backaway speed
        // R3 - range as which we start to backaway
        R1 = 124; R2 = 180; R3 = 325;
      } else {
        Serial.println(F("> MainTask_TrackWall()"));
        // R1 - range below which we slow down
        // R2 - range which detemines max backaway speed
        // R3 - range as which we start to backaway
        // R4 - range at which we stop going forward
        // R5 - range at which we reach max speed going forward
        // R6 - range at which we start going forward
        R1 = 89; R2 = 117; R3 = 189; R4 = 225; R5 = 298; R6 = 325;
      } SubTask++; break;
    case 1:
      // depending on zMode set a walk direction
      if (LedMode != 5) {SetLedMode(5);}
      if (zMode == 0) {
        // start expecting to go backward
        StartWalk(-3);
      } else {
        // start expecting to go forward
        StartWalk(-1);
      }
      SubTask = 2; break;
    case 2: // wait for a wall to appear and if necessary change direction
      // only move onto SubTask 3 when a wall appears
      walkInterval = 60000; 
      if (LedMode != 5) {SetLedMode(5);}
//      Serial.println(Walk);
      if ((Range >= R0) && (Range <= R3)) {
        SubTask = 3; Walk= abs(Walk);
        if (Walk != 3) {
          // change walk direction from forwards to backwards
          Walk = 0;  WalkNext = 3; SetLegStates(); // set leg heights
          SubTask = 99;
        }
      } else if ((Range >= R4) && (Range <= R6)) {
        SubTask = 3; Walk= abs(Walk);
        if (Walk != 1) {
          // change walk direction from backards to forwards
          Walk = 0; WalkNext = 1; SetLegStates(); // set leg heights
          SubTask = 99;
        }
      } break;
    case 3: // scan for wall and respond
      if (LedMode != 5) {SetLedMode(5);}
      if (Range >= R0) {
        if (Range > RM) {
          // target moved out of range so stop
          SubTask = 4;
        } else if ((zMode > 0) && (Range > R6)) {
          // if walking forward let it continue, but if not start; both moderately
          if (abs(Walk) == 1) {Walk = 1;}
          else {StartWalk(1);} walkInterval = 20000;
        } else if ((zMode > 0) && (Range >= R4) && (Range <= R6)) {
          // far wall detected so walking speed depends on how close it is
          walkInterval = map(Range, R4,R5,1,12);
          walkInterval = 60000/walkInterval;
          if (Range >= R5) {walkInterval = 5000;} // cap minimum interval
          GetSteeringDrive(1);
          Walk = 1;
        } else if ((Range <= R3) && (Range >= R1)) {
          // close wall detected so backaway speed depends on how close it is
          walkInterval = map(Range,R3,R2,1,12);
          walkInterval = 60000/walkInterval;
          if (Range <= R2) {walkInterval = 5000;} // cap minimum interval
          GetSteeringDrive(-1);
          Walk = 3;
        } else if (Range < R1) {
          // too close wall detected so slow down backaway speed
          walkInterval = map(Range, R0,R1,1,12);
          walkInterval = 60000/walkInterval;
          GetSteeringDrive(-1);
          Walk = 3;
        } else {
          // in centre deadband so stop
          SubTask = 4;}
      } else {
        // no wall detected but one had been previously
        SubTask = 4;
      } break;
    case 4: // stop moving and put feet down and wait
      Walk = -abs(Walk); walkInterval = 20000;
      MemStore(); // store current values
// servo        (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type         ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr,Cnt)
      GoToAngles(LegDn,   -1,LegDn,   -1,   -1,LegDn,   -1,LegDn,LegDn,   -1,LegDn,   -1,  1); // put all feet down immediately
      SubTask++; anyMilli = millis(); break;  // start timer for pause period
    case 5: // wait for a valid range then return to previous movement
      if (LedMode != 5) {SetLedMode(5);}
      if ((Range >= R0) && (Range <= R3)) {
        // valid reverse range
        if (Walk == -3) {
          // previously walking backwards so raise lowered legs and continue
          MemRead(); TgtCnt = 1; SubTask = 3;
        } else {SubTask = 2;} // walking direction changing
      } else if ((zMode > 0) && (Range >= R4) && (Range <= R6)) {
        // valid forward range
        if (Walk == -1) {
          // previously walking forwards so lower legs and continue
          MemRead(); TgtCnt = 1; SubTask = 3;
        } else {SubTask = 2;} // walking direction changing
      }
      if (zMode == 0) {
        // zMode = 0 back away from an approaching target
        // check after a delay to see if a return movement is needed
        if ((millis() - anyMilli) >= 5000) {
          if (Walked < -16) {SubTask++;} // at least half a stride backwards
          else {anyMilli = millis();}
        }
      } else if (zMode == 1) {
        // zMode = 1 maintains a fixed distance from target
        // if stood still after a delay return to ready state
        if ((millis() - anyMilli) >= 5000) {
          anyMilli = millis(); StartWalk(Walk); SubTask = 2;
        }
      }
      break;
    case 6:
      // prepare for return to start movement
      WalkLftDrive = 128; WalkRgtDrive = 128;
      walkInterval = 20000;
      Walk = 0; SubTask++; break;
    case 7:
      // change direction of walk movement to forward
      walkInterval = 40000; SetLedMode(21);
      SetLegsPrevious(); Walk = 1;
      SubTask++; break;
    case 8: 
      // check reached beginning or seen another wall?
      Serial.println("Walked = " + String(Walked));
      if (((Range >= R0) && (Range <= R3)) || (Walked >= -1)) {
        // stop walking forward
        Walk = 0; SubTask = 1; SetLedMode(5);
// servo          (    0,    1,    2,    3,    4,    5,    6,    7,    8,    9,   10,   11) servo reference
// type           ( Claw,Shldr, Claw,Shldr,Shldr, Claw,Shldr, Claw, Claw,Shldr, Claw,Shldr,Cnt)
        GoToAngles(LegDn,   -1,LegDn,   -1,   -1,LegDn,   -1,LegDn,LegDn,   -1,LegDn,   -1,  1);
      }
      // if approaching start point slow down
      if (Walked > -100) {if (walkInterval < 40000) {walkInterval += 200;}}
      else if (walkInterval > 10000) {walkInterval -= 100;}
      break;
    case 99:
      // one cycle to allow leg heights to be set before moving
      Walk = WalkNext; SubTask = 3; break;
  } MainRun = false;
}

// --------------------------------------------------------------------------------

void MoveLegToAngle(int16_t zS,int16_t zA0,int16_t zL,int16_t zA1) {
  // move shoulder zS to specific servo angle zA0 by raising and lowering leg zL
  // valid shoulder zS values are 1,3,4,6,9,11
  // valid leg zL values are 0,2,5,7,8,10 
  if (ESC) return; // exit if this flag is set
  
  if (!ServoEn) {attachServos(0);}
  anyDel = 120;
  // front right leg Shldr1,Claw0
  SetAngle(20 + zS,zA0); // get the shoulder target angle in micros, but don't move it
  if (servoVal[zS] != servoTgt[zS]) {
    SetAngle(zL,LegUp); delayLoop(anyDel); SetAngle(zS,zA0); delayLoop(anyDel);
  } SetAngle(zL,zA1);

  if (autoOFF >= 0) {autoOFF = 5;} // set auto-OFF timeout
}

// --------------------------------------------------------------------------------

void MoveNow() {
  // called to move legs to a GoTo position, ignoring all other inputs
  // returns to calling function once TgtCnt == 0
  while (TgtCnt > 0) {Leg_Engine(); delay(20);}
  synchLoopTimers(0); // re-synch loop timers
}

// --------------------------------------------------------------------------------

void MoveToAngles(int16_t zA0,int16_t zA1,int16_t zA2,int16_t zA3,int16_t zA4,int16_t zA5,int16_t zA6,int16_t zA7,int16_t zA8,int16_t zA9,int16_t zAA,int16_t zAB) {
  // move to specific servo angles by raising and lowering legs
  // process starts with all legs being put down first, then each leg is adjusted in turn
  // working from front right in a clockwise direction
  if (ESC) return; // exit if this flag is set
  
  if (!ServoEn) {attachServos(0);}
  anyDel = 120;
// servo    (      0,     1,      2,     3,     4,      5,     6,      7,      8,     9,     10,    11) servo reference
// type     (  Claw0,Shldr1,  Claw2,Shldr3,Shldr4,  Claw5,Shldr6,  Claw7,  Claw8,Shldr9,  ClawA,ShldrB)
  GoToAngles(  LegDn,    -1,  LegDn,    -1,    -1,  LegDn,    -1,  LegDn,  LegDn,    -1,  LegDn,    -1,3);
  // front right leg Shldr1,Claw0
  SetAngle(21,zA1); // get the shoulder target angle in micros, but don't move it
  if (servoVal[1] != servoTgt[1]) {
    SetAngle(0,LegUp); delayLoop(anyDel); SetAngle(1,zA1); delayLoop(anyDel);
  } SetAngle(0,zA0);
  // centre right leg Shldr3,Claw2
  SetAngle(23,zA3); // get the shoulder target angle in micros, but don't move it
  if (servoVal[3] != servoTgt[3]) {
    SetAngle(2,LegUp); delayLoop(anyDel); SetAngle(3,zA3); delayLoop(anyDel);
  } SetAngle(2,zA2);
  // rear right leg Shldr4,Claw5
  SetAngle(24,zA4); // get the shoulder target angle in micros, but don't move it
  if (servoVal[4] != servoTgt[4]) {
    SetAngle(5,LegUp); delayLoop(anyDel); SetAngle(4,zA4); delayLoop(anyDel);
  } SetAngle(5,zA5);
  // rear left leg Shldr6,Claw7
  SetAngle(26,zA6); // get the shoulder target angle in micros, but don't move it
  if (servoVal[6] != servoTgt[6]) {
    SetAngle(7,LegUp); delayLoop(anyDel); SetAngle(6,zA6); delayLoop(anyDel);
  } SetAngle(7,zA7);
  // centre left leg Shldr9,Claw8
  SetAngle(29,zA9); // get the shoulder target angle in micros, but don't move it
  if (servoVal[9] != servoTgt[9]) {
    SetAngle(8,LegUp); delayLoop(anyDel); SetAngle(9,zA9); delayLoop(anyDel);
  } SetAngle(8,zA8);
  // front left leg ShldrB,ClawA
  SetAngle(31,zAB); // get the shoulder target angle in micros, but don't move it
  if (servoVal[11] != servoTgt[11]) {
    SetAngle(10,LegUp); delayLoop(anyDel); SetAngle(11,zAB); delayLoop(anyDel);
  } SetAngle(10,zAA); delayLoop(anyDel);

  if (autoOFF >= 0) {autoOFF = 5;} // set auto-OFF timeout
}

// --------------------------------------------------------------------------------

void NeutralTurnLeft(int16_t zAng, unsigned long zInterval) {
  // neutral turn to the left by a given angle, and speed
  // any range values are ignored
  unsigned long zwalkInterval = walkInterval; // save current walk period
  walkInterval = zInterval;
  StartWalk(5);
  SetLedMode(25);
  loopWhileTurn(zAng); Walk = 0;
  walkInterval = zwalkInterval;
}

// --------------------------------------------------------------------------------

void NeutralTurnLeftEcho(int16_t zAng, unsigned long zInterval) {
  // neutral turn to the left by a given angle, and speed
  // note that speed can also be set by walkInterval loop period
  // the actual angle turned is approximately (zAng * 28)/45
  // if zAng < 0 then don't move legs to the drop position, assume already there
  unsigned long zwalkInterval = walkInterval; // save current walk period
  walkInterval = zInterval;
  StartWalk(5);
  SetLedMode(25);
  loopWhileTurnEcho(zAng); Walk = 0;
  walkInterval = zwalkInterval;
}

// --------------------------------------------------------------------------------

void NeutralTurnRight(int16_t zAng, unsigned long zInterval) {
  // neutral turn to the right by a given angle, step and speed
  // any range values are ignored
  unsigned long zwalkInterval = walkInterval; // save current walk period
  walkInterval = zInterval;
  StartWalk(6);
  SetLedMode(26);
  loopWhileTurn(zAng); Walk = 0;
  walkInterval = zwalkInterval;
}

// --------------------------------------------------------------------------------

void NeutralTurnRightEcho(int16_t zAng, unsigned long zInterval) {
  // neutral turn to the right by a given angle, step and speed
  // any range values are ignored
  unsigned long zwalkInterval = walkInterval; // save current walk period
  walkInterval = zInterval;
  StartWalk(6);
  SetLedMode(26);
  loopWhileTurnEcho(zAng); Walk = 0;
  walkInterval = zwalkInterval;
}

// --------------------------------------------------------------------------------

void SetHeadTgtCnt(int zAngDif) {
  // sets the movement count value based on the displacement AngDif
  HeadTgtCnt = (16 * abs(zAngDif))/AngDifMax;
  if (HeadTgtCnt < 1) {HeadTgtCnt = 1;}
}

// --------------------------------------------------------------------------------

void SetLegsDown() {
  // stores the leg positions then puts them all down on the floor
  MemStore();
  SetAngle(0,LegDn); SetAngle(2,LegDn); SetAngle(5,LegDn);
  SetAngle(7,LegDn); SetAngle(8,LegDn); SetAngle(10,LegDn);
}

// --------------------------------------------------------------------------------

void SetLegStates() {
  // set leg heights depending on WalkCnt, used when reversing direction
//  LegSt0 = LegDn; LegSt2 = LegDn; LegSt5 = LegDn; LegSt7 = LegDn; LegSt8 = LegDn; LegSt10 = LegDn;
//  if ((WalkCnt > 1) && (WalkCnt < 44)) {LegSt3 = LegUp; LegSt7 = LegUp;}
//  else if (WalkCnt > 45) {LegSt1 = LegUp; LegSt5 = LegUp;}
//  GoToAngles(-1,LegSt1,-1,LegSt3,-1,LegSt5,-1,LegSt7,1);
}

// --------------------------------------------------------------------------------

void SetLegsPrevious() {
  // recover stored leg positions
  MemRead(); SetToTgtNow();
}

// --------------------------------------------------------------------------------

void SetWalkMax() {
  // set the value of WalkMax based on WalkSpeed
  // think of WalkSpeed as the gear you are in
  switch(WalkSpeed) {
    case 1: WalkMax = 16; break;
    case 2: WalkMax = 28; break;
    case 3: WalkMax = 40; break;
    case 4: WalkMax = 52; break;
    case 5: WalkMax = 60; break;
  }
}

// --------------------------------------------------------------------------------

void StartWalk(int zWalk) {
  // initiates one of four walk directions
  // if zWalk is -ve then position feet but don't walk
  // 0 - stationary
  // 1 - walk forward
  // 2 - walk to the right
  // 3 - walk backwards
  // 4 - walk to the left
  // 5 - neutral turn to the left
  // 6 - neutral turn to the right
  Walk = 0; // stop walking
  switch (abs(zWalk)) {
//    case 0: MoveToAngles(111,LegDn,112,LegDn,114,LegDn,113,LegDn); break;
    case 1: // walk forward
      LegDn = LegDwnMin; LegUp = LegUpMin;
      WL_C = 90; WR_C = 90; WL_I = 1; WR_I = -2;
      // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
      // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
      MoveToAngles(LegDn,WR_C,LegDn,WR_C,WR_C,LegDn,WL_C,LegDn,LegDn,WL_C,LegDn,WL_C);
      // the angle range (Walk_U - Walk_L) must always be divisible by 2
      Walk_L = 74; Walk_U = 106;
      break;
//    case 2: MoveToAngles(67,LegDn,69,LegDn,68,LegDn,66,LegDn); break;
    case 3: // walk backwards
      LegDn = LegDwnMin; LegUp = LegUpMin;
      WL_C = 90; WR_C = 90; WL_I = 1; WR_I = -2;
      // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
      // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
      MoveToAngles(LegDn,WR_C,LegDn,WR_C,WR_C,LegDn,WL_C,LegDn,LegDn,WL_C,LegDn,WL_C);
      // the angle range (Walk_U - Walk_L) must always be divisible by 2
      Walk_L = 74; Walk_U = 106;
      break;
//    case 4: MoveToAngles(68,LegDn,66,LegDn,67,LegDn,69,LegDn); break;
    case 5: // neutral turn left
      LegDn = LegDwnMin; LegUp = LegUpMin;
      WL_C = 90; WR_C = 90; WL_I = 1; WR_I = -2;
      // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
      // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
      MoveToAngles(LegDn,WR_C,LegDn,WR_C,WR_C,LegDn,WL_C,LegDn,LegDn,WL_C,LegDn,WL_C);
      // the angle range (Walk_U - Walk_L) must always be divisible by 2
      Walk_L = 74; Walk_U = 106;
      break;
    case 6: // neutral turn right
      LegDn = LegDwnMin; LegUp = LegUpMin;
      WL_C = 90; WR_C = 90; WL_I = 1; WR_I = -2;
      // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
      // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
      MoveToAngles(LegDn,WR_C,LegDn,WR_C,WR_C,LegDn,WL_C,LegDn,LegDn,WL_C,LegDn,WL_C);
      // the angle range (Walk_U - Walk_L) must always be divisible by 2
      Walk_L = 74; Walk_U = 106;
      break;
  }
  Walk = zWalk; // start walking if Walk > 0
}

// ---------------------------------------------------------------------------------

void TestMoveServos() {
  // in TEST mode set servos to 1500 µs by default
  // SW0 toggles between 1500 µs and a short sweeping motion left/right/left
  // SW0 long press cycles through calibrated values
  // if serial port becomes active then abort this process
  // if WiFi is active then that overides SW0 settings
  if (!ServoEn) {attachServos(10); attachHeadServo(); synchLoopTimers(0);}

  if (SerialRx) {return;} // abort movement
  bool zOp = false;
  if (WiFiRxRec) {
    // connected to WiFi so respond to the Nunchuk Y-axis
    // we don't move far in case of collisions with other servos
    ServoAny = 1500 + ((JoyY - 128) * 1); zOp = true;
  } else {
    // servo movements are controlled by a set of tasks
    switch (TestTask) {
      case 1:
        // Normal SW0 modes, static or small sweep
        Testmode = 1; Walk = 0; // ensure walk engine is OFF
        if (ServoInc == 0) {
          if (ServoAny == 1500) {return;} // do nothing
          ServoAny = 1500;                // move to mid-point value
        } else if (ServoInc > 0) {
          ServoAny += 5; if (ServoAny >= 1600) {ServoInc = -1;}
        } else {
          ServoAny -= 5; if (ServoAny <= 1400) {ServoInc =  1;}
        }  zOp = true; break;
      
      case 2:
        // move legs through calibrated values, starting from rest
        Testmode = 2; Walk = 0; // ensure walk engine is OFF
        // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
        // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
        GoToAngles(  69,  90,  69,  90,  90,  69,  90,  69,  69,  90,  69,  90,-50);
        HeadToAngle(90,50);
        TestNext = TestTask + 1; TestTask = 99; break;
      case 3:
        // centred raise legs to 90°
        GoToAngles(  90,  -1,  90,  -1,  -1,  90,  -1,  90,  90,  -1,  90,  -1,-50);
        HeadToAngle(20,50);
        TestNext = TestTask + 1; TestTask = 99; break;
      case 4:
        // move legs to forward
        GoToAngles(  -1,  21,  -1,  30,  68,  -1,  68,  -1,  -1,  30,  -1,  21,-50);
        HeadToAngle(90,50);
        TestNext = TestTask + 1; TestTask = 99; break;
      case 5:
        // centred raise legs to 180°
        GoToAngles( 223,  90, 223,  90,  90, 223,  90, 223, 223,  90, 223,  90,-50);
        HeadToAngle(160,50);
        TestNext = TestTask + 1; TestTask = 99; break;
      case 6:
        // centred raise legs to 180°
        GoToAngles(  90,  -1,  90,  -1,  -1,  90,  -1,  90,  90,  -1,  90,  -1,-50);
        HeadToAngle(90,50);
        TestNext = TestTask + 1; TestTask = 99; break;
      case 7:
        // legs back raised to 223°
        GoToAngles(  -1, 120,  -1, 150, 180,  -1, 180,  -1,  -1, 150,  -1, 120,-50);
        HeadToAngle(90,50);
        TestNext = 2; TestTask = 99; break;

      case 10:
        // demonstrate walking sequence
        Testmode = 10; HeadToAngle(90,50); Walk = 98; TestTask++; break;
      case 11: break; // do nothing whilst walking
      
      case 99:
        // wait for movement to complete
        if (TgtCnt == 0) {TestTask = TestNext;}
        break;
    }
  }
  if (zOp) {
    for (int zS = 0;zS <= 12;zS++) {
      servoInst[zS].writeMicroseconds(ServoAny); servoVal[zS] = ServoAny;
    }
    servoHead.writeMicroseconds(ServoAny); HeadVal = ServoAny;
//    Serial.println(ServoAny);

  }
}

// --------------------------------------------------------------------------------

void Walk_Engine() {
  //###############################################################################
  //
  //  Walk Engine tasks
  //
  //###############################################################################
  // cald from the main loop if Walk > 0
  // 0 - stationary
  // 1 - walk forward
  // 2 - NA
  // 3 - walk backwards
  // 4 - NA
  // 5 - neutral turn to the left
  // 6 - neutral turn to the right
  if (autoOFF >= 0) {autoOFF = 5;} // set auto-OFF timeout
  switch(Walk) {
    case 1:
      //###########################################################################
      //
      //  Walk forward with steering
      //
      //###########################################################################
      // apply steering
      WL_CM = (((WL_C - 90) * WalkLftDrive)/128) + 90;
      WR_CM = (((WR_C - 90) * WalkRgtDrive)/128) + 90;
      // left front/rear pair + right centre
      if (WL_I > 0) {
        // moving legs backward at normal speed whilst applying steering
        // set the shoulder angles before putting the legs down
        SetAngle( 6,WL_CM); SetAngle(11,WL_CM);
        SetAngle( 3,(((WL_C - 90) * WalkRgtDrive)/128) + 90);
        SetAngle( 7,LegDn); SetAngle( 2,LegDn); SetAngle(10,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 7,LegUp); SetAngle( 2,LegUp); SetAngle(10,LegUp);
        SetAngle( 6,WL_CM); SetAngle(11,WL_CM);
        SetAngle( 3,(((WL_C - 90) * WalkRgtDrive)/128) + 90);
      }
      // right front/rear pair + left centre
      if (WR_I > 0) {
        // moving legs backward at normal speed whilst applying steering
        // set the shoulder angles before putting the legs down
        SetAngle( 4,WR_CM); SetAngle( 1,WR_CM);
        SetAngle( 9,(((WR_C - 90) * WalkLftDrive)/128) + 90);
        SetAngle( 5,LegDn); SetAngle( 8,LegDn); SetAngle( 0,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 5,LegUp); SetAngle( 8,LegUp); SetAngle( 0,LegUp);
        SetAngle( 4,WR_CM); SetAngle( 1,WR_CM);
        SetAngle( 9,(((WR_C - 90) * WalkLftDrive)/128) + 90);
      }
      // adjust counters and change increments at angle limits
      Walked++; // count up clocks in forward direction
      WL_C+= WL_I;
      if (WL_C >= Walk_U) {WL_I = -2;}
      if (WL_C <= Walk_L) {WL_I =  1;}
      WR_C+= WR_I;
      if (WR_C >= Walk_U) {WR_I = -2;}
      if (WR_C <= Walk_L) {WR_I =  1;}
      break;

    case 3:
      //###########################################################################
      //
      //  Walk backward with steering
      //
      //###########################################################################
      // apply steering
      WL_CM = (((WL_C - 90) * WalkLftDrive)/128) + 90;
      WR_CM = (((WR_C - 90) * WalkRgtDrive)/128) + 90;
      // left front/rear pair + right centre
      if (WL_I > 0) {
        // moving legs backward at normal speed whilst applying steering
        // set the shoulder angles before putting the legs down
        SetAngle( 6,WL_CM); SetAngle(11,WL_CM);
        SetAngle( 3,(((WL_C - 90) * WalkRgtDrive)/128) + 90);
        SetAngle( 7,LegDn); SetAngle( 2,LegDn); SetAngle(10,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 7,LegUp); SetAngle( 2,LegUp); SetAngle(10,LegUp);
        SetAngle( 6,WL_CM); SetAngle(11,WL_CM);
        SetAngle( 3,(((WL_C - 90) * WalkRgtDrive)/128) + 90);
      }
      // right front/rear pair + left centre
      if (WR_I > 0) {
        // moving legs backward at normal speed
        // set the shoulder angles before putting the legs down
        SetAngle( 4,WR_CM); SetAngle( 1,WR_CM);
        SetAngle( 9,(((WR_C - 90) * WalkLftDrive)/128) + 90);
        SetAngle( 5,LegDn); SetAngle( 8,LegDn); SetAngle( 0,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 5,LegUp); SetAngle( 8,LegUp); SetAngle( 0,LegUp);
        SetAngle( 4,WR_CM); SetAngle( 1,WR_CM);
        SetAngle( 9,(((WR_C - 90) * WalkLftDrive)/128) + 90);
      }
      // adjust counters and change increments at angle limits
      Walked--; // count down clocks in backward direction
      WL_C-= WL_I;
      if (WL_C >= Walk_U) {WL_I =  1;}
      if (WL_C <= Walk_L) {WL_I = -2;}
      WR_C-= WR_I;
      if (WR_C >= Walk_U) {WR_I =  1;}
      if (WR_C <= Walk_L) {WR_I = -2;}
      break;

    case 5:
      //###########################################################################
      //
      //  Neutral turn left
      //
      //###########################################################################
     // left front/rear pair + right centre
      if (WL_I > 0) {
        // moving legs backward at normal speed whilst applying steering
        // set the shoulder angles before putting the legs down
        SetAngle( 6,WL_C); SetAngle(11,WL_C);
        SetAngle( 3,180 - WL_C);
        SetAngle( 7,LegDn); SetAngle( 2,LegDn); SetAngle(10,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 7,LegUp); SetAngle( 2,LegUp); SetAngle(10,LegUp);
        SetAngle( 6,WL_C); SetAngle(11,WL_C);
        SetAngle( 3,180 - WL_C);
      }
      // right front/rear pair + left centre
      if (WR_I > 0) {
        // moving legs backward at normal speed
        // set the shoulder angles before putting the legs down
        SetAngle( 4,WR_C); SetAngle( 1,WR_C);
        SetAngle( 9,180 - WR_C);
        SetAngle( 5,LegDn); SetAngle( 8,LegDn); SetAngle( 0,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 5,LegUp); SetAngle( 8,LegUp); SetAngle( 0,LegUp);
        SetAngle( 4,WR_C); SetAngle( 1,WR_C);
        SetAngle( 9,180 - WR_C);
      }
      // adjust counters and change increments at angle limits
      Walked++; // count up clocks whilst turning
  Serial.println("Walked=" + String(Walked));
      WL_C-= WL_I;
      if (WL_C >= Walk_U) {WL_I =  1;}
      if (WL_C <= Walk_L) {WL_I = -2;}
      WR_C+= WR_I;
      if (WR_C >= Walk_U) {WR_I = -2;}
      if (WR_C <= Walk_L) {WR_I =  1;}
      break;

    case 6:
      //###########################################################################
      //
      //  Neutral turn right
      //
      //###########################################################################
      // left front/rear pair + right centre
      if (WL_I > 0) {
        // moving legs backward at normal speed whilst applying steering
        // set the shoulder angles before putting the legs down
        SetAngle( 6,WL_C); SetAngle(11,WL_C);
        SetAngle( 3,180 - WL_C);
        SetAngle( 7,LegDn); SetAngle( 2,LegDn); SetAngle(10,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 7,LegUp); SetAngle( 2,LegUp); SetAngle(10,LegUp);
        SetAngle( 6,WL_C); SetAngle(11,WL_C);
        SetAngle( 3,180 - WL_C);
      }
      // right front/rear pair + left centre
      if (WR_I > 0) {
        // moving legs backward at normal speed
        // set the shoulder angles before putting the legs down
        SetAngle( 4,WR_C); SetAngle( 1,WR_C);
        SetAngle( 9,180 - WR_C);
        SetAngle( 5,LegDn); SetAngle( 8,LegDn); SetAngle( 0,LegDn);
      } else {
        // moving legs forward at double speed
        // lift the claws before changing the shoulder angles
        SetAngle( 5,LegUp); SetAngle( 8,LegUp); SetAngle( 0,LegUp);
        SetAngle( 4,WR_C); SetAngle( 1,WR_C);
        SetAngle( 9,180 - WR_C);
      }
      // adjust counters and change increments at angle limits
      Walked++; // count up clocks whilst turning
  Serial.println("Walked=" + String(Walked));
      WL_C+= WL_I;
      if (WL_C >= Walk_U) {WL_I = -2;}
      if (WL_C <= Walk_L) {WL_I =  1;}
      WR_C-= WR_I;
      if (WR_C >= Walk_U) {WR_I =  1;}
      if (WR_C <= Walk_L) {WR_I = -2;}
      break;
      
    case 98:
      //###########################################################################
      //
      //  Keyboard driven walk engine
      //
      //###########################################################################
      // special walk evaluation mode from serial cmd values
      // initialise walk sequence
      W_Dn = 110; W_Up = 120; WL_C = Walk_S; WR_C = Walk_S; WL_I = 1; WR_I = -2;
      // servo  (   0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11) servo reference
      // type   (Clw0,Shd1,Clw2,Shd3,Shd4,Clw5,Shd6,Clw7,Clw8,Shd9,ClwA,ShdB)
      GoToAngles(W_Dn,WR_C,W_Dn,WR_C,WR_C,W_Dn,WL_C,W_Dn,W_Dn,WL_C,W_Dn,WL_C,-100);
      MoveNow();
      // the angle range (Walk_U - Walk_L) must always be divisible by 2
      Walk_L = 74; Walk_U = 106; Walk = 99; ReportWalkVals();
      break;
    case 99:
      // run walk sequence
      // left front/rear pair + right centre
      if (WL_I > 0) {
        // moving legs backward at normal speed
        SetAngle( 7,W_Dn); SetAngle( 2,W_Dn); SetAngle(10,W_Dn);
        SetAngle( 6,WL_C); SetAngle( 3,WL_C); SetAngle(11,WL_C);
      } else {
        // moving legs forward at double speed
        SetAngle( 7,W_Up); SetAngle( 2,W_Up); SetAngle(10,W_Up);
        SetAngle( 6,WL_C); SetAngle( 3,WL_C); SetAngle(11,WL_C);
      }
      // right front/rear pair + left centre
      if (WR_I > 0) {
        // moving legs backward at normal speed
        SetAngle( 5,W_Dn); SetAngle( 8,W_Dn); SetAngle( 0,W_Dn);
        SetAngle( 4,WR_C); SetAngle( 9,WR_C); SetAngle( 1,WR_C);
      } else {
        // moving legs forward at double speed
        SetAngle( 5,W_Up); SetAngle( 8,W_Up); SetAngle( 0,W_Up);
        SetAngle( 4,WR_C); SetAngle( 9,WR_C); SetAngle( 1,WR_C);
      }
      // adjust counters and change increments at angle limits
      WL_C+= WL_I;
      if (WL_C >= Walk_U) {WL_I = -2;}
      if (WL_C <= Walk_L) {WL_I =  1;}
      WR_C+= WR_I;
      if (WR_C >= Walk_U) {WR_I = -2;}
      if (WR_C <= Walk_L) {WR_I =  1;}
      break;
  }
}

// --------------------------------------------------------------------------------
