
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  disply.init()         - initialise the display
  disply.invertDisplay()- inverted display mode
  disply.end()          - free memory used by the display
  disply.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  disply.normalDisplay()- normal display mode
  disply.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // if not in hold mode then switch between displays once every 2s
  if (DispDel > 0) {DispDel--; return;} // hold off delay if > 0
  if (DispClr) {display.clear(); DispClr = false; return;}
  if (DispDisp) {display.display(); DispDisp = false; DispClr = true; return;}

  // if not on hold we switch display modes every time DispNxt == 0
  if (!DispHold) {
    DispNxt--; if (DispNxt < 1) {
      DispNxt = 50; // reset 2 sec timer
      DispPnt++; if (DispPnt > 3) {DispPnt = 0;}
      DispCnt = 1;  // create next display immediately
    }
  }

  DispCnt--;
  if (DispCnt < 1) {
    switch(DispPnt) {
      case 0: DispCnt = 10; Display_Battery(); break;
      case 1: DispCnt = 10; Display_Status(); break;
      case 2: DispCnt = 5; Display_I2C(); break;
      case 3: DispCnt = 5; Display_Joystick(); break;
    }
    display.display(); DispDisp = false; DispClr = true;

    // the following line puts a rectangle around every display, showing the boundry region
//    display.drawRect(0,0,128,64);
  }
}

// --------------------------------------------------------------------------------

void Display_Battery() {
  // display battery information
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 06, "Battery");
  if (BatAvg > BatUSB) {
    // battery powered mode
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 27, String(BatAvg));
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, String(float(BatAvg)/BatCal) + "v");
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
    if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
    if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
    display.drawString(127, 27, String(AnyLong) + "%");
    // display the time
    display.setTextAlignment(TEXT_ALIGN_LEFT);
    display.drawString(0, 48, getHmsTime(millis()));
    display.setTextAlignment(TEXT_ALIGN_RIGHT);
    display.drawString(127, 48, BatTime$);
    display.display();
  } else {
    // USB powered mode
    display.setTextAlignment(TEXT_ALIGN_CENTER);
    display.drawString(64, 27, "USB Pwr");
    // display the time
    display.drawString(64, 48, getHmsTime(millis()));
  }
}

// --------------------------------------------------------------------------------

void Display_Exhausted() {
  // display the word EXHAUSTED due to critifcal battery failure
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, "EXHAUSTED");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 40, "Recharge Now!");
  display.display();
}

// --------------------------------------------------------------------------------

void Display_I2C() {
  // display 6 bytes received from Wii I2C
  // display I2C values as decimal 0 - 255
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  display.setFont(ArialMT_Plain_10); DispCX = 58;
  display.drawString(DispCX,  0, String(received_data[0]));
  display.drawString(DispCX, 10, String(received_data[1]));
  display.drawString(DispCX, 20, String(received_data[2]));
  display.drawString(DispCX, 30, String(received_data[3]));
  display.drawString(DispCX, 40, String(received_data[4]));
  display.drawString(DispCX, 50, String(received_data[5]));
  // display I2C values as binary 00000000 - 11111111
  display.setTextAlignment(TEXT_ALIGN_LEFT); DispCX = 66;
  display.drawString(DispCX,  0, GetBIN(received_data[0]));
  display.drawString(DispCX, 10, GetBIN(received_data[1]));
  display.drawString(DispCX, 20, GetBIN(received_data[2]));
  display.drawString(DispCX, 30, GetBIN(received_data[3]));
  display.drawString(DispCX, 40, GetBIN(received_data[4]));
  display.drawString(DispCX, 50, GetBIN(received_data[5]));
  // display 'I2C' in top left corner
  display.setFont(ArialMT_Plain_16);
  display.drawString(2,  0, "Wii");
       if (WiiType == 'N') {display.drawString(2, 18, "Nnk");}
  else if (WiiType == 'C') {display.drawString(2, 18, "Csc");}
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 6, "Wii Transceiver");
  display.setFont(ArialMT_Plain_10);
  display.drawString(64, 28, "v0.01");
  display.drawString(64, 50, myMAC);
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Held() {
  // called when locking a display mode, puts LOCKED in centre of screen but does
  // not clear the reast of the contents
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  DispCX = display.getStringWidth("LOCKED");
  DispCY = 2 + (DispCX/2);
  display.setColor(BLACK);
  display.fillRect(64 - DispCY,21,DispCX + 5,25);
  display.setColor(WHITE);
  display.drawRect(64 - DispCY,21,DispCX + 5,25);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 20, "LOCKED");
  display.display();
  DispClr = true; // initiate a clear screen for the enxdt update
}

// --------------------------------------------------------------------------------

void Display_Joystick() {
  // display joystick valus and offsets
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 4, "Joysticks");

  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.drawString(16,24, "X = ");
  if (Wii_Connected) {
           if (WiiType == 'N') {
             display.drawString(44,24, String(received_data[0]));
             display.drawString(86,24, String(JoyXOff));
    } else if (WiiType == 'C') {
      display.drawString(44,24, String(received_data[0] & 63));
      display.drawString(86,24, "---");
    } else {
      display.drawString(44,24, "---");
      display.drawString(86,24, "---");
    }
  } else {
    display.drawString(44,24, "---");
    display.drawString(86,24, "---");
  }

  display.drawString(16,46, "Y = ");
  if (Wii_Connected) {
           if (WiiType == 'N') {
             display.drawString(44,46, String(received_data[1]));
             display.drawString(86,46, String(JoyYOff));
    } else if (WiiType == 'C') {
      display.drawString(44,46, String(received_data[1] & 63));
      display.drawString(86,46, "---");
    } else {
      display.drawString(44, 46, "---");
      display.drawString(86, 46, "---");
    }
  } else {
    display.drawString(44, 46, "---");
    display.drawString(86, 46, "---");
  }
}

// --------------------------------------------------------------------------------

void Display_Reboot() {
  // called when a soft reset is about to be initiated
  display.clear();
  display.setFont(ArialMT_Plain_24);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 10, "RESET");
  display.setFont(ArialMT_Plain_16);
  display.drawString(64, 40, "Pending");
  display.display();
}

// --------------------------------------------------------------------------------

void Display_Status() {
  // display status information
//  display.drawRect(0,0,128,64);
  display.setFont(ArialMT_Plain_16);  // fonts are ArialMT_Plain_10, ArilMT_Plain_24
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.drawString(64, 4, "WiFi STATUS");
  
  if (Connected) {
    display.setFont(ArialMT_Plain_16);
    display.drawString(64, 24, "Connected");
    display.drawString(64, 44, WiFi_Dev);
  } else {
    display.setFont(ArialMT_Plain_16);
    display.drawString(64, 24, "Disconnected");
    display.drawString(64, 44, "Listening");
  }
}

// --------------------------------------------------------------------------------
