
// --------------------------------------------------------------------------------

void BatteryLow() {
  // code comes here if battery voltage drops below critical limit
  // it never exits this function
  if (!BatPwr) {return;}  // started on USB power so ignore low voltage readings
  
  Display_Exhausted();
  Serial.println(F("Critical voltage shutdown"));
  STOP = true;
  delay(60000); // display EXHAUSTED for 1 minute then clear the display
  display.clear(); display.display();
}

// --------------------------------------------------------------------------------

void ClearI2CData() {
  // whenh Wii device is not connected we set the Rx buffer data to defaults
  // this assumed Nunchuk data
  received_data[0] = 128;
  received_data[1] = 128;
  received_data[2] = 0;
  received_data[3] = 0;
  received_data[4] = 0;
  received_data[5] = 3;
}

// --------------------------------------------------------------------------------

void clearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void clearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void getRemoteName() {
  // called from the receive function when in the disconnected state. It is
  // assumed that the remote device will be establishing a connection using its
  // own name, so extract that from the Rx buffer
  WiFi_Dev = "";
  for (int zP = 0; zP < Rx_len; zP++) {
      WiFi_Dev += Rx_Buff.ESPdata[zP];
  }
  // clear Rx buffer for next Rx
  clearRxBuff(Rx_len);
}

// --------------------------------------------------------------------------------

void Flush_Serial() {
  // empty the contents of the incoming serial port
  while (Serial.available() > 0) {Tx_Char = Serial.read();} Tx_Char = 0;
}

// --------------------------------------------------------------------------------

String GetBIN(byte zVal) {
  // returns an 8 character binary representation of zVal
  String zBIN$ = "";
  // convert zVal into binary
  if (zVal & 0B10000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B01000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00100000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00010000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00001000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000100) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000010) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000001) {zBIN$ += "1";} else {zBIN$ += "0";}
  return zBIN$;
}

// --------------------------------------------------------------------------------

String getHmTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min base on zMillis
  String zTime$ = ""; zMillis = zMillis/60000;  // reduce zMillis to minutes
  long zU = zMillis/60; zMillis = zMillis%60;    // get hours and minutes
  zTime$ = String(zU) + ":";
  if (zMillis < 10) {zTime$ += "0";}
  zTime$ += String(zMillis);  // add minutes
  return zTime$;
}

// --------------------------------------------------------------------------------

String getHmsTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min:sec base on zMillis
  String zTime$ = "";
  long zU = zMillis/3600000; zMillis = zMillis%3600000;
  if (zU > 0) {zTime$ = String(zU) + ":";}
  zU = zMillis/60000; zMillis = zMillis%60000;
  if (zU < 10) {zTime$ += "0";}
  zTime$ += String(zU);
  zU = zMillis/1000;
  if (zU < 10) {zTime$ += ":0";} else {zTime$ += ":";}
  zTime$ += String(zU);
  return zTime$;
}

// --------------------------------------------------------------------------------

void PackI2CData() {
  // copy the Nunchuk data into the 8 byte transmit buffer with a checksum byte
  // use first byte as a header to define the type of data
  Tx_Buff.ESPdata[0] = WiiType; 
  Tx_Buff.ESPdata[1] = received_data[0];
  Tx_Buff.ESPdata[2] = received_data[1];
  Tx_Buff.ESPdata[3] = received_data[2];
  Tx_Buff.ESPdata[4] = received_data[3];
  Tx_Buff.ESPdata[5] = received_data[4];
  Tx_Buff.ESPdata[6] = received_data[5];
  // calculate checksum byte for Wii data only, ignore WiiType
  Tx_Buff.ESPdata[7] = (received_data[0] ^ received_data[1] ^ received_data[2] ^ received_data[3] ^ received_data[4] ^ received_data[5]);
  Tx_len = 8; // adjust the pointer for potential serial data to be added
}

// --------------------------------------------------------------------------------

void ReadBattery() {
  // called every 40ms from the main loop
  // read the battery voltage and calculate an average  
  BatVol = analogRead(BatVolPin); BatTms = millis(); // take the reading
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/20;     // ADC is averaged over 20 readings to remove noise
  
  //#############################################################################
  // Life prediction code
  //#############################################################################
  // BatCnt is set to 10 second interval to start with, to ignore initial readings
  if (BatDel > 0) {BatDel--;}
  else if (BatT0 == 0) {
    // this is the first measurement after 10 second start delay
    // this values are the cornerstone of the prediction
    BatT0 = BatTms; BatV0 = BatAvg; BatV1 = BatAvg;
  } else {
    if (BatAvg < BatV1) {
      // there has been a change in the battery voltage setting a new discharge point
      // recalculate battery life predicition base on new value and time
      BatV1 = BatAvg; // record the latest low point
      AnyLong = ((BatTms - BatT0)*(BatV1 - BatCritical))/(BatV0 - BatV1);
  //      Serial.print("AnyLong "); Serial.println(AnyLong);
      if (AnyLong > 0) {
        BatTime$ = getHmTime(AnyLong);
      } else {BatTime$ = "--:--";}
      BatDel = 0; // reset 1 second delay timer
    }
  }  
  if (!BatPwr) {return;}  // on USB power so nothing to display
  
//  Serial.println("BatAvg = " + String(BatAvg) + " == " + String(float(BatAvg)/BatCal) + " V");
  // test for critical volatge
  if (BatAvg <= BatCritical) {BatteryLow();}
}

// --------------------------------------------------------------------------------

void ReadI2CCalData() {
  // read a block of data from the Nunchuk calibration registers, defined as:
  /*
    1   0G value of X-axis [9:2]
    2   0G value of Y-axis [9:2]
    3   0G value of Z-axis [9:2]
    4   LSB of Zero value of X,Y,Z axes
    5   1G value of X-axis [9:2]
    6   1G value of Y-axis [9:2]
    7   1G value of Z-axis [9:2]
    8   LSB of 1G value of X,Y,Z axes
    9   Joystick X-axis maximum
    10  Joystick X-axis minimum
    11  Joystick X-axis center
    12  Joystick Y-axis maximum
    13  Joystick Y-axis minimum
    14  Joystick Y-axis center
    15  Checksum
    16  Checksum

  */
  Wire.beginTransmission(nunchuk_address); //Start communication with the Nunchuck.
  Wire.write(0x20);                        //We want to start reading at register (20 hex)
  WiiError = Wire.endTransmission();       //End the transmission and check ACK
  delay(1);
  //
  // check to ensure that the Nunchuk has responded
  if (WiiError == 0) {
    // no error reported so read the sixteen Nunchuk registers
    Wire.requestFrom(nunchuk_address,16);      //Request 16 bytes from the Nunchuck
    for(int i = 0; i < 16; i++) received_data[i] = Wire.read(); //Copy the bytes to the received_data array

    if (reportEn > 0) {
      Serial.println(F("\nWii Calibration Data:")); // send title
      ReportNunchukCalData();
    }
  }
}

// --------------------------------------------------------------------------------

void ReadI2CData() {
  // read a block of data from the Nunchuk controller
  Wire.beginTransmission(nunchuk_address); //Start communication with the Nunchuck.
  Wire.write(0x00);                        //We want to start reading at register (00 hex)
  WiiError = Wire.endTransmission();       //End the transmission and check ACK
  delay(1);
  //
  // check to ensure that the Nunchuk has responded
  if (WiiError == 0) {
    // no error reported so read the six Nunchuk registers
    Wire.requestFrom(nunchuk_address,6);      //Request 6 bytes from the Nunchuck
    for(byte i = 0; i < 6; i++) received_data[i] = Wire.read(); //Copy the bytes to the received_data array
//    received_data[5] = received_data[5] & 0x03; // mask off accelerometer bits
    // check data validity in case Nunchuk has crashed. Under such cercumstances
    // all data registers will appear to hold a value of 255
    // we ignore the joystick registers as they may actually be at 255
    if ((received_data[2] == 255) && (received_data[3] == 255) && (received_data[4] == 255)) {
      WiiError = 1;
    } else {
      // Nunchuk appears to be working correctly
      // we will use the 4th reading as being the centre value
      // both Nunchuk and Classic are mapped from 0 - 255
      if (WiiCal < 5) {
        WiiCal++;
        if ((WiiCal == 4) && (WiiType == 'N')) {
          JoyXOff = received_data[0];
          JoyYOff = received_data[1];
          XYOffAdj = true;
        }
      }
      if (XYOffAdj) {
        // XY offset adjustment enabled, so re-map received values
        // this only applies to a Nunchuk controller
        if (received_data[0] <= JoyXOff) {
          received_data[0] = map(received_data[0],0,JoyXOff,0,128);
        } else {received_data[0] = map(received_data[0],JoyXOff,255,128,255);}
        if (received_data[1] <= JoyYOff) {
          received_data[1] = map(received_data[1],0,JoyYOff,0,128);
        } else {received_data[1] = map(received_data[1],JoyYOff,255,128,255);}
      }
    }
  }
  if (WiiError != 0) {
    // we re-initialise the Nunchuk if a failure in the I2C comms occurs
    // if it fails to reinitialise we flag the failure
    WiiInitialise();
    if (WiiErrCnt > WiiErrRet) {
      if (Wii_Connected) {Serial.println(F("Wii disconnected!"));}
      // clear all of the data
      ClearI2CData(); Wii_Connected = false; WiiType = '-';
    }
  }
}

// --------------------------------------------------------------------------------

void ReadI2CIdent() {
  // read a block of data from the Nunchuk extension register
  Wire.beginTransmission(nunchuk_address); //Start communication with the Nunchuck.
  Wire.write(0xFA);                        //We want to start reading at register (FA hex)
  WiiError = Wire.endTransmission();       //End the transmission and check ACK
  delay(1);
  //
  // check to ensure that the Nunchuk has responded
  if (WiiError == 0) {
    // no error reported so read the six Nunchuk registers
    WiiType = 0;  // clear the current type flag incase controllers are swapped out
    Wire.requestFrom(nunchuk_address,6);      //Request 6 bytes from the Nunchuck
    for(int i = 5; i >=0; i--) received_data[i] = Wire.read(); //Copy the bytes to the received_data array
    // check data validity in case Nunchuk has crashed. Under such cercumstances
    // all data registers will appear to hold a value of 255
    // we ignore the joystick registers as they may actually be at 255
    WiiType = '-';  // Wii device unknown
    if ((received_data[2] == 255) && (received_data[3] == 255) && (received_data[4] == 255)) {
      WiiError = 1;
    } else {
      // looks ok so determine the controller type
      // 0xA4200000 - Nunchuk
      // 0xA4200101 - Classic Controller
      if ((received_data[5] == 0) && (received_data[4] == 0) && (received_data[3] == 164) && (received_data[2] == 32) && (received_data[1] == 0) && (received_data[0] == 0)) {
        // Nunchuck controller - 0xA4200000
        Serial.println(F("\nWii Nunchuk connected")); // send title
        WiiType = 'N';
      }
      else if ((received_data[5] == 0) && (received_data[4] == 0) && (received_data[3] == 164) && (received_data[2] == 32) && (received_data[1] == 1) && (received_data[0] == 1)) {
        // Classic controller - 0xA4200101
        Serial.println(F("\nWii Classic connected")); // send title
        WiiType = 'C';
      }
    }
  }
  if (WiiError == 0) {
    // Wii device appears to have responded correctly so report device ident
    if (reportEn > 0) {
      Serial.println(F("\nWii Ident Data:")); // send title
      ReportNunchukData();
    }
  }
}

// --------------------------------------------------------------------------------

void readSW1() {
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  sw1State = digitalRead(D4);
  if (sw1State == LOW) {
    //###########################################
    // SW1 button is pressed down
    //###########################################
    if (DispCnt < 10) {DispCnt = 10;} // don't allow display to change when button is pressed
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
      Serial.println("SW1 Hi-LO");
      sw1Cnt++;     // count on the falling edge
      sw1Timer = 0; // reset the timer so we can count SW! pressed time
    }
    if (sw1Timer == 60) {
      // button held down for 600ms so lock the display in this mode
      DispHold = true; Display_Held();  // lock to this mode
    } else if (sw1Timer >= 200) {
      // button held down for >= 2 seconds so force a soft reset
      Serial.println("Reset pending..."); Display_Reboot();
      Reboot = true;  // stop all main loop tasks
      resetFunc();    //call reset
    }
  } else {
    //###########################################
    // sw1State == HIGH as SW1 button is not pressed
    //###########################################
    if (sw1LastState == LOW) {
      // SW1 has just been released
      if ((sw1Timer > 0) && (sw1Timer < 60)) {
        // short press so increment the display
        DispNxt = 0; DispHold = false; sw1Cnt = 0; sw1Timer = 0;
      }
    }
    if (sw1Timer >= 100) {
      // button released for 1 sec so assume valid button count
      // we can put code here which responds to button counts
      sw1Cnt = 0; sw1Timer = 0;
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// --------------------------------------------------------------------------------

void ReportNunchukCalData() {
  // send nunchuk calibration data to the serial port, with formatting
  for(byte i = 0; i < 16; i++) {ReportNunchukDataLn(i);}
}

// --------------------------------------------------------------------------------

void ReportNunchukData() {
  // send nunchuk data to the serial port, with formatting
  for(byte i = 0; i < 6; i++) {ReportNunchukDataLn(i);}
}

// --------------------------------------------------------------------------------

void ReportNunchukDataLn(int zP) {
  // reports a single line for a given register zP, range 0 - 15
  Serial.print(F("Bite ")); Serial.print(zP); Serial.print(F(" = "));
  if (received_data[zP] < 100) Serial.print(F(" "));
  if (received_data[zP] < 10) Serial.print(F(" "));
  Serial.print(received_data[zP]);       //output I2C data from nunchuk
  Serial.print(F("\t"));
  // this section packs out the binary field to a constant 8 characters
  if (received_data[zP] < 128) Serial.print(F("0"));
  if (received_data[zP] < 64) Serial.print(F("0"));
  if (received_data[zP] < 32) Serial.print(F("0"));
  if (received_data[zP] < 16) Serial.print(F("0"));
  if (received_data[zP] < 8) Serial.print(F("0"));
  if (received_data[zP] < 4) Serial.print(F("0"));
  if (received_data[zP] < 2) Serial.print(F("0"));
  Serial.println(received_data[zP],BIN);       //output I2C data from nunchuk
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void Sync_Timers() {
  // called whenever a process delay could compromise the loop timers
  // the timers are pre-loaded with a time phase to avoid overlap
  next2ms = millis() + 2;  // prime 2ms loop timer for serial data
  next10ms = millis() + 4; // primt 10ms loop timer
  next40ms = millis() + 6; // prime 40ms loop timer 5ms out of phase
}

// --------------------------------------------------------------------------------

void TxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  Tx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Tx_Chk = Tx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// --------------------------------------------------------------------------------

void WiiInitialise() {
  // innitialise the Wii over the I2C bus
  WiiError = 1; WiiErrCnt = 0; WiiTask = 0;
  while (WiiError != 0) {
    // wait in this loop initialising the Wii until it responds
    Wire.begin();                             //Start the I2C as master
    delay(1);                                 //Short delay
    Wire.beginTransmission(nunchuk_address);  //Start communication with the Nunchuck
    Wire.write((byte)0xF0);                   //We want to write to register (F0 hex)
    Wire.write((byte)0x55);                   //Set the register bits as 01010101
    Wire.endTransmission();                   //End the transmission
    delay(1);                                 //Short delay
    Wire.beginTransmission(nunchuk_address);  //Start communication with the Nunchuck
    Wire.write((byte)0xFB);                   //We want to write to register (FB hex)
    Wire.write((byte)0x00);                   //Set the register bits as 00000000
    WiiError = Wire.endTransmission();        //End the transmission
    delay(1);                                 //Short delay
    WiiErrCnt++;                              //Track the number of attempts
    if (WiiErrCnt > WiiErrRet) {break;}       // no Nunchuk detected
  }
  // if called from the main loop we need to reset the timers
  Sync_Timers();
}

// --------------------------------------------------------------------------------

void WiiInitialiseTask() {
  // innitialise the Wii over the I2C bus as a series of tasks when not connected
  // this function is called every 10ms
  switch (WiiTask) {
    case 0:
      Wire.begin();                             //Start the I2C as master
      break;
    case 1:
      Wire.beginTransmission(nunchuk_address);  //Start communication with the Nunchuck
      Wire.write((byte)0xF0);                   //We want to write to register (F0 hex)
      Wire.write((byte)0x55);                   //Set the register bits as 01010101
      Wire.endTransmission();                   //End the transmission
      break;
    case 2:
      Wire.beginTransmission(nunchuk_address);  //Start communication with the Nunchuck
      Wire.write((byte)0xFB);                   //We want to write to register (FB hex)
      Wire.write((byte)0x00);                   //Set the register bits as 00000000
      WiiError = Wire.endTransmission();        //End the transmission
      if (WiiError == 0) {
        Wii_Connected = true; Serial.println("Wii initialised");
        ReadI2CIdent();
        XYOffAdj = false; WiiCal = 0; ReadI2CData();
      } else {WiiType = '-'; ClearI2CData(); Wii_Connected = false;} 
      break;
  }
}

// --------------------------------------------------------------------------------
